//------------------------------------------------------------------------------
//  Author:    	MXF Team
//  Comment:	MXF_TK: This class defines the timecode interface.
//
//  The following convention is used for defining parameters and retrun values:
//      in : input parameter (will be used for computation in the function)
//      out : output parameter (will be modified by the function call)
//      in-out : input parameter that may be modified by the function call.
//      free : MXFTk user is reponsible for the deletion of this parameter
//             unless it is passed as a parameter of another function without 
//             the "free" tag.
//
// Copyright (c) 2008 OPENCUBE. All rights reserved.
// -----------------------------------------------------------------------------
#ifndef OC_I_TIMECODE_HPP
#define OC_I_TIMECODE_HPP

#include <mxf_tk/mxf_tk.hpp>

namespace opencube
{
  namespace MXFTK_NAMESPACE
  {
    /* I_timecode:
     * This class is used to represent a timecode.
     */
    class I_timecode
    {
    public:
      /*
       * Displays timecode. in a XX:XX:XX:XX format.
       * Returns NULL if an error occured.
       */
      virtual const char* CALLING_CONVENTION get_time() const=0;

      /* Get the hours */
      virtual uint8_t CALLING_CONVENTION get_hour() const=0;

      /* Get the minutes */
      virtual uint8_t CALLING_CONVENTION get_minute() const=0;

      /* Get the seconds */
      virtual uint8_t CALLING_CONVENTION get_second() const=0;

      /* Gets the frames */
      virtual uint8_t CALLING_CONVENTION get_frame() const=0;

      /* Get the timecode as a number of frames elapsed from 00:00:00:00*/
      virtual uint64_t CALLING_CONVENTION get_frame_count()=0;

      /*
       * Jumps to the next timecode frame.
       * Returns true if successful
       */
      virtual bool CALLING_CONVENTION next()=0;

      /*
       * Jumps to the previous timecode frame.
       * Returns true if successful
       */
      virtual bool CALLING_CONVENTION previous()=0;
	
      /*
       * Jumps to an amount of frames forward.
       * Returns true if successful
       */
      virtual bool CALLING_CONVENTION add(uint64_t)=0;

      /*
       * Jumps to an amount of frames backward.
       * Returns true if successful
       */
      virtual bool CALLING_CONVENTION sub(uint64_t)=0;

      /*
       * Returns the timecode's frame rate.
       * Returns 0 if an error occured.
       */
      virtual uint16_t CALLING_CONVENTION get_frame_rate() const=0;

      /*
       * Is this timecode a drop frame or not?
       */
      virtual bool CALLING_CONVENTION get_drop_frame() const=0;

      /*
       * Is this timecode strictly greater than the supplied one?
       */
      virtual bool CALLING_CONVENTION greater_than(I_timecode* /* in free */) const=0;

      /*
       * Is this timecode strictly smaller than the supplied one?
       */
      virtual bool CALLING_CONVENTION smaller_than(I_timecode* /* in free */) const=0;

      /*
       * Is this timecode equal to the supplied one?
       */
      virtual bool CALLING_CONVENTION equal_to(I_timecode* /* in free */) const=0;
    };

    /*
     * Retrieves a timecode interface.
     * I_timecode** pointer to the timecode to be created.
     * const char* timecode value (for instance "00:08:15:04")
     * uint16_t denotes the frame rate.
     * bool specifies if a drop frame timecode format must be activated.
     */
    extern "C" mxf_tk_API bool CALLING_CONVENTION NewTimecodeInterface(I_timecode** /* out free */, 
                                                                       const char* /* in free */, 
                                                                       uint16_t /* in */,
                                                                       bool /* in */);

    /*
     * Retrieves a timecode interface.
     * I_timecode** pointer to the timecode to be created.
     * The four first uint16_t denoted hour, minute, second, frame
     * The last uint16_t denotes the frame rate.
     * bool specifies if a drop frame timecode format must be activated.
     */
    extern "C" mxf_tk_API bool CALLING_CONVENTION NewTimecodeInterfaceByValue(I_timecode** /* out free */, 
                                                                              uint16_t /* in */, 
                                                                              uint16_t /* in */, 
                                                                              uint16_t /* in */, 
                                                                              uint16_t /* in */, 
                                                                              uint16_t /* in */, 
                                                                              bool /* in */);

    /*
     * Retrieves a copy of a timecode interface.
     */
    extern "C" mxf_tk_API bool CALLING_CONVENTION CopyTimecodeInterface(I_timecode** /* out free */, 
                                                                        const I_timecode* /* in free */);

    /*
     * Frees a timecode interface.
     */
    extern "C" mxf_tk_API bool CALLING_CONVENTION FreeTimecodeInterface(I_timecode** /* in out */);
  }
}

#endif
