//------------------------------------------------------------------------------
//  Author:    	MXF Team
//  Comment:	MXF_TK: This class defines the track interface.
//
//  The following convention is used for defining parameters and retrun values:
//      in : input parameter (will be used for computation in the function)
//      out : output parameter (will be modified by the function call)
//      in-out : input parameter that may be modified by the function call.
//      free : MXFTk user is reponsible for the deletion of this parameter
//             unless it is passed as a parameter of another function without 
//             the "free" tag.
//
// Copyright (c) 2008 OPENCUBE. All rights reserved.
// -----------------------------------------------------------------------------
#ifndef OC_I_TRACK_HPP
#define OC_I_TRACK_HPP

#include <mxf_tk/mxf_tk.hpp>
#include <mxf_tk/I_track_item.hpp>

#define INT64_ERROR   0x7fffffffffffffffLL
#define UINT64_ERROR  0xffffffffffffffffULL

namespace opencube
{
  namespace MXFTK_NAMESPACE
  {
    /* I_track:
     * A class describing a track from a material.
     * This can be an essence or a descriptive metadata track.
     * No essence data can be read from it (this can be performed on an I_concrete_track).
     */
    class I_track
    {
    public:
      /*
       * Returns type of this track.
       * Returns track_type_error if an error occured.
       */
      virtual track_type CALLING_CONVENTION type() const=0;

      /*
       * Returns type of this track in a string. Returned pointer musn't
       * be deleted.
       * Returns NULL if an error occured.
       */
      virtual const char* CALLING_CONVENTION type_name() const=0;

      /*
       * Returns the edit rate of this track. The edit rate is
       * the number of edit units per second. Returned pointer 
       * must be freed using FreeRationalInterface.
       * Returns NULL if an error occured.
       */
      virtual rational* /* free */ CALLING_CONVENTION edit_rate() =0;

      /*
       * Returns the origin of this track measured in edit units.
       * Returns 0x7fffffffffffffff if an error occured.
       */
      virtual int64_t CALLING_CONVENTION origin() =0;

      /*
       * Returns the duration of this track measured in edit units.
       * Returns -1 if unknown.
       */
      virtual int64_t CALLING_CONVENTION duration() =0;

      /* 
       * Returns an id uniquely identifying this track within this material
       * Returns 0xffffffff if an error occured.
       */
      virtual uint32_t CALLING_CONVENTION get_track_id() const=0;

      /*
       * Sets the name of this track (utf16)
       * the size_type is used to specify the size in bytes of this string.
       * API user can free the wchar_t pointer after calling this function.
       * Returns true if successful
       */
      virtual bool CALLING_CONVENTION set_track_name(const wchar_t* /* in free */, 
                                                     size_t /* in */)=0;

      /*
       * Gets the name of this track (utf16) and retrieves its size.
       * The string returned will be destroyed upon deletion of the I_track*.
       * Returns NULL if an error occured.
       */
      virtual const wchar_t* CALLING_CONVENTION get_track_name(size_t& /* out */) const=0;

      /*
       * Returns an ordered list of track items. MXF_TK user should 
       * try to dynamically cast each track_item into a source clip, 
       * dm_source_clip or dm_segment. Returned list must always
       * be deleted using free_elements.
       * Returns NULL if an error occured.
       */
      virtual ordered_track_item_list* /* free */ CALLING_CONVENTION elements() =0;

      /*
       * Frees list returned by elements().
       * Returns true if successful
       */
      virtual bool CALLING_CONVENTION free_elements(ordered_track_item_list* /* in */) const=0;

      /*
       * This function will operate only if the current track is
       * a metadata track. It appends a dm_segment to the track.
       * If the current track is :
       * - timeline metadata : the dm_segment is added at the end of the
       * last track_item of this track. Its duration is set in the
       * metadata material.
       * - event metadata : Position and duration of the dm_segment on
       * this track are set in the metadata_material.
       * - static track : no timing properties. Metadata is global to 
       * the material.
       * Returned items must not be deleted nor edited.
       * Track list is a list of documented tracks.
       * Returns true if successful.
       */
      virtual bool CALLING_CONVENTION add_metadata(I_metadata_material* /* in */, 
                                                   track_list* /* in */)=0;

      /*
       * This function will operate only if the current track is
       * a static or event metadata track. It removes a dm segment (metadata material)
       * from the current track. 
       * Note this function will be inoperative on a timeline metadata track. 
       */
      virtual bool CALLING_CONVENTION remove_metadata(I_metadata_material* /* in */)=0;

      /* 
       * Seeks a given edit unit position and returns the track items at this position.
       * This item must be casted into the appropriate type. Note that if it's an
       * I_source_clip, the referenced track will be set ready for reading with the
       * read() function.
       * The int64_t parameter is the position along the track. Its value will
       * be added to the origin of this track. 
       * Therefore a track with an origin = 5 and a duration = 25 can be accessed
       * with a int64_t position comprised between -5 and 20.
       * If the seek operation cannot be performed (seeking beyond track
       * scope) returned value is NULL.
       */
      virtual ordered_track_item_list* /* free */ CALLING_CONVENTION item_seek_eu(int64_t /* in */)=0;

      /* 
       * Seeks a given timecode and returns the track items at this position.
       * This item must be casted into the appropriate type. Note that if it's an
       * I_source_clip, the referenced track will be set ready for reading with the
       * read() function.
       * The int64_t parameter is the position along the track. Its value will
       * be added to the origin of this track. 
       * Therefore a track with an origin = 5 and a duration = 25 can be accessed
       * with a int64_t position comprised between -5 and 20.
       * If the seek operation cannot be performed (seeking beyond track
       * scope) returned value is NULL.
       */
      virtual ordered_track_item_list* /* free */ CALLING_CONVENTION item_seek_timecode(const I_timecode* /* in free */)=0;

      /*
       * These functions return the start and end timecode for this track.
       * Returns NULL if it cannot be computed.
       * API user is responsible for the deletion of the I_timecode* returned.
       */
      virtual I_timecode* /* free */ CALLING_CONVENTION get_start_timecode()=0;
      virtual I_timecode* /* free */ CALLING_CONVENTION get_end_timecode()=0;

      /*
       * Utility functions
       * Convert a timecode value into an edit unit offset of this track and vice versa.
       * Returns -1 if the edit unit offset cannot be computed or NULL if the timecode
       * value cannot be computed.
       */
      virtual int64_t CALLING_CONVENTION timecode_to_edit_unit(const I_timecode* /* in free */)=0;
      virtual I_timecode* /* free */ CALLING_CONVENTION edit_unit_to_timecode(int64_t)=0;
    };
  }
}

#endif
