//------------------------------------------------------------------------------
//  Author:    	Thierry FRANZETTI
//  Comment:	MXF_TK: This example illustrates DCI OpAtom mxf file creation.
//
// Copyright (c) 2008 OPENCUBE. All rights reserved.
// -----------------------------------------------------------------------------
#if defined(BC6)	// Borland C Builder 6.0
#include <vcl.h>
#endif

#include <mxf_tk/mxf_tk.hpp>
#include <mxf_tk/I_timecode_material.hpp>
#include <mxf_tk/I_concrete_material.hpp>
#include <mxf_tk/I_dcp1_file.hpp>

#include <iostream>
#include <string>
#include <list>

using namespace opencube::MXFTK_NAMESPACE;
using namespace std;

// MxfTk error handler
I_mxf_error_handler* err_h = NULL;

//Throws an exception when an error occured 
void throw_on_error(); 

//Called if an error occured.
void display_errors();

int main(int argc, char* argv[])
{
  if (argc != 4 )
  {
    cout << "usage : out.mxf source file crypting" << endl;
    cout << "\tcrypting id \"true\" or \"false\"" << endl;
    exit(0);
  }

  string do_crypt_str(argv[3]);
  if(do_crypt_str != string("true") && do_crypt_str != string("false"))
  {
    cout << "usage : out.mxf source_file crypting" << endl;
    cout << "\tcrypting id \"true\" or \"false\"" << endl;
    exit(0);
  }

  bool do_crypt = (do_crypt_str == string("true"));


  // the opatom mxf file we are going to write.
  I_dcp1_file* pi_opatom_file = NULL;
  // the interface to manipulate the source to be wrapped.
  I_concrete_material* pi_conc_mat = NULL;
  // this boolean is set to true if pi_conc_mat needs to be destroyed
  bool delete_conc_mat = false;
  // Path to the source files to be wrapped
  crypted_locators* pi_cloc = NULL; // encryption case
  locators* pi_loc = NULL; // plaintext case (no encryption)

  rational * edit_rate = NULL;

  try
  {
    wcout << L"Retrieve MxfTk Error Handler" << endl;
    if (!GetMxfErrorHandlerInterface(&err_h))
      throw 8001;

    wcout << L"MxfTk Initialization" << endl;
    InitMXF_TK();
    throw_on_error();

    wcout << L"Create the opatom mxf file interface" << endl;
    NewDcp1FileInterface(&pi_opatom_file, argv[1]);
    throw_on_error();

    if(!NewRationalInterface(&edit_rate, 24, 1))
    {
      throw 8002;
    }

    if(do_crypt)
    {
      wcout << L"Build source files list" << endl;
      GetEmptyCryptedLocatorInterface(&pi_cloc);
      throw_on_error();
      
      crypted_locator_element * elem = NULL;
      
      uint32_t plaintext_off(0);
      
      uint8_t* cryptographic_key_id = new uint8_t[16];
      uint8_t* cipher_key = new uint8_t[16];
      
      for(unsigned int ii = 0; ii < 16; ii++)
      {
	cryptographic_key_id[ii] = ii;
	cipher_key[ii] = 15 - ii;
      }
      
      NewCryptedLocatorElementInterface(&elem, argv[2], plaintext_off, 
					cryptographic_key_id, cipher_key);
      
      pi_cloc->add(&elem);

      delete[] cryptographic_key_id;
      delete[] cipher_key;

      wcout << L"Create source material" << endl;
      NewCryptedMaterialInterface(&pi_conc_mat, pi_cloc, dcp1_wrapping, edit_rate);
      throw_on_error();
      delete_conc_mat = true;
      
      FreeCryptedLocatorElementInterface(&elem);
    }
    else
    {
      wcout << L"Build source files list" << endl;
      GetEmptyLocatorInterface(&pi_loc);
      throw_on_error();

      pi_loc->add((const char**)&(argv[2]));
      
      wcout << L"Create source material" << endl;
      NewConcreteMaterialInterface(&pi_conc_mat, pi_loc, dcp1_wrapping, edit_rate);

      throw_on_error();
      delete_conc_mat = true;
    }
    
    wcout << L"Compute concrete material duration" << endl;
    pi_conc_mat->compute_duration();

    wcout << L"Attach material to the mxf file" << endl;
    pi_opatom_file->set_material(pi_conc_mat);
    throw_on_error();
    delete_conc_mat = false;  
    // delete_conc_mat set to false now coz pi_conc_mat will be destroyed 
    // upon deletion of pi_opatom_file (FreeOpAtomFileInterface).

    wcout << L"Write mxf file on disk" << endl;
    pi_opatom_file->flush();
    throw_on_error();

    wcout << L"Wait termination" << endl;
    pi_opatom_file->wait_end_flush_thread();
    throw_on_error();
  }
  catch (int e)
  {
    if(e == 8001)
    {
      wcout << "Unable to retrieve error handler" << std::endl;
    }
    else if(e == 8002)
    {
      wcout << "Unable to build rational edit rate" << std::endl;
    }
    else
    {
      wcout << "An error occured!" << std::endl;
      // Retrieving details of the error(s) and display it(them).
      display_errors();
    }
  }

  wcout << L"Close Process" << endl;
  if (pi_opatom_file)
  {
    FreeDcp1FileInterface(&pi_opatom_file);
  }
  if (pi_conc_mat && delete_conc_mat)
  {
    FreeConcreteMaterialInterface(&pi_conc_mat);
  }
  if (pi_cloc)
  {
    FreeCryptedLocatorInterface(&pi_cloc);
  }
  if (pi_loc)
  {
    FreeLocatorInterface(&pi_loc);
  }
  if (edit_rate)
  {
    FreeRationalInterface(&edit_rate);
  }

  return 0;
}

void throw_on_error()
{
  if (err_h->has_an_error())
    throw 1;
}

void display_errors()
{
  // Retrieve error handler
  if (!err_h)
    return;

  // Retrieve list of errors
  error_list* err_l = err_h->get_errors();
  if (err_l)
  {
    err_l->begin();
    I_mxf_error** ppi_mxf_error = NULL;
    while ((ppi_mxf_error = err_l->next()) != NULL) //go through the list
    {
      I_mxf_error* pi_error = *ppi_mxf_error;
      // and display each error 
      wcout << L"Error " << (int)pi_error->get_id() << L": " << pi_error->get_error_msg() << endl;
    }
    // free error list interface
    err_h->free_errors(err_l);
  }
}
