//------------------------------------------------------------------------------
//  Author:    	Thierry FRANZETTI
//  Comment:	MXF_TK: An example to read the content of an mxf stream.
//  This example simulates a possible implementation for reading data
//  from a linear storage device.
//  You should be familiar with the non-streaming interfaces
//  before going through this example. This is an advanced example(!)
//
// Copyright (c) 2008 OPENCUBE. All rights reserved.
// -----------------------------------------------------------------------------
#if defined(BC6)	// Borland C Builder 6.0
#include <vcl.h>
#endif

#include <mxf_tk/I_mxf_error.hpp>
#include <mxf_tk/I_mxf_file.hpp>
#include <mxf_tk/I_metadata_material.hpp>

#include <iostream>
#include <sstream>
#include <map>

using namespace opencube;
using namespace std;

// MxfTk error handler
I_mxf_error_handler* err_h = NULL;

//Throws an exception when an error occured 
void throw_on_error(); 

//Called if an error occured.
void display_errors();

bool closed_and_complete = false;

// This is the size of the buffer where the mxf stream is read 
// (buffer from the mxf linear-storage device).
#define BUF_SIZE 16384

// We derive the pure virtual class I_input_mxf_stream_task.
// This class provides several functions that will be automatically 
// called by MxfTk whenever some data is ready to be read or
// whenever some data is required from the mxf stream.
// For the conveniency of the demonstration, this implementation reads
// mxf data from a file however the same mechanism could
// be applied to a true streaming device.
class I_user_input_mxf_stream_task : public I_input_mxf_stream_task
{
public:
  // Constructor -> mxf file (mxf streaming device buffer)
  I_user_input_mxf_stream_task(std::string);

  // Destructor -> close file (free streaming device buffer)
  virtual ~I_user_input_mxf_stream_task();

  // Sets the size of MxfTk internal buffer used to bufferize the
  // mxf stream. MxfTk user is free to change its value.
  // Increasing the buffer size may help to speed up the process
  // Here, it is set to the same size as the streaming device buffer (BUF_SIZE). 
  virtual size_t CALLING_CONVENTION get_buffer_size() { return BUF_SIZE; };

  // MxfTk will call this function whenever it requires soma data from the
  // mxf stream. User should provide an implementation "feeding" Mxftk
  // with the content of the storage device's buffer.
  virtual void CALLING_CONVENTION data_request(I_mxf_file*, size_t);

  // do nothing here -> we are not interested in the essence
  virtual void CALLING_CONVENTION data_to_read(I_generic_material*, I_track*, I_source_clip*, uint64_t)
  {

  }

  // MxfTk will call this function when a seek operation is required on 
  // the input mxf stream. This call will occur only if a seek operation
  // was requested by the API user (thanks to seek_timecode). Therefore
  // it is perfectly valid to have a nil implementation for this function 
  // if you do not wish to perform seek or if your storage device is not
  // capable of seeking.
  virtual uint64_t CALLING_CONVENTION seekpos_request(uint64_t);

  // Returns the size of the MXF file being streeamed (mandatory)
  virtual uint64_t CALLING_CONVENTION stream_size() { return _stream_size; }

  // Notifies MXFTk that it can seek on the stream (mandatory)
  virtual bool CALLING_CONVENTION seekable() { return true; }

  // Called by MXFTk to notify the user that some data must be written on the stream.
  virtual void CALLING_CONVENTION updated_data(const uint8_t* , size_t);

  // This function is called whenever a new header metadata is met in the file.
  // Previous mxf_files become obsolete.
  virtual bool CALLING_CONVENTION updated_mxf_file(I_mxf_file*, bool closed, bool complete, uint64_t);
protected:
  // storage device buffer
  uint8_t* buffer;
  // file handle (storage device)
  FILE* file;
  // filename 
  std::string _mxf_file_name;
  // its size
  uint64_t _stream_size;

  // A map containing the names of the output files
  // There will be one file for each output track of
  // the generic material. You may replace the output 
  // files by any other storage mechanism.
  std::map <uint32_t, FILE*> out_files;
};

// Constructor : Open the mxf storage device for reading (file) 
// and retrieve its buffer.
I_user_input_mxf_stream_task::I_user_input_mxf_stream_task(std::string mxf_file_name)
: I_input_mxf_stream_task()
{
  file = fopen(mxf_file_name.c_str(), "r+b");  // open storage device
  if (!file || ferror(file))
    wcout << L"An error occured while opening the mxf file" << endl;

  // we need to compute the stream size
  if (fseek(file, 0, SEEK_END)!=0)
  {
    wcout << L"Failed to seek at the end of the file" << endl;
  }

  _stream_size = ftell(file);
  // get back to the beginning
  fseek(file, 0, SEEK_SET);
    

  buffer = new uint8_t[BUF_SIZE]; // retrieve storage device's buffer
  _mxf_file_name = mxf_file_name; // store for later use.
}

// Destructor : Frees storage device's buffer and close
// output files.
I_user_input_mxf_stream_task::~I_user_input_mxf_stream_task()
{
  delete [] buffer;
  fclose(file);

  for (map<uint32_t, FILE*>::iterator it = out_files.begin(); it != out_files.end(); it++)
  {
    FILE* f= it->second;
    fclose(f);
  }
}

// MxfTk notifies its need for data from the input mxf stream.
// Called with an interface on the mxf file being processed and
// ask for a given number of bytes.
void CALLING_CONVENTION I_user_input_mxf_stream_task::data_request(I_mxf_file* pi_mxf_file, size_t s_req)
{
  try
  {
    // Check MxfTk doesn't ask more than the storage device's buffer.
    if (s_req>BUF_SIZE)
      s_req = BUF_SIZE;

    // It is perfectly valid to give less data than the amount specified (s_req)
    // data_request() is called whenever MxfTk internal buffer is empty (all
    // the previous mxf data has been processed)

    // read data from the storage device
    // If no data is available, it is possible to use mutex to
    // stop the current process as long as no data is available from
    // the storage device.
    size_t s_read = fread(buffer, sizeof(uint8_t), s_req, file);
    if (ferror(file)) throw 1;

    // There's some data in the buffer.
    if (s_read>size_t(0))
    {
      // Then give it to MxfTk
      size_t written = pi_mxf_file->write(buffer, s_read);
      wcout << L"Mxf Stream: " << (long)s_read << " bytes written" << endl;
      if (written<s_read) throw 1; // couldn't write all  
    }
    // No more data to be processed. The file is closed then notifies it to MxfTk
    else
    {
      pi_mxf_file->end_of_stream();
      wcout << L"Mxf Stream: closed" << endl; 
    }
    throw_on_error();
  }
  catch(int)
  {
    wcout << L"An error occured!" << endl;
    // Retrieving details of the error(s) and display it(them).
    display_errors();
  }
}

// The user asked for seeking a given timecode
// Notifies the storgae device it has to seek to a given position
// If seek timecode always returns NULL you do not need to
// provide an implementation for this function.
uint64_t CALLING_CONVENTION I_user_input_mxf_stream_task::seekpos_request(uint64_t pos)
{
  //Possible implementation
  if (fseek(file, long(pos), SEEK_SET)!=0)
  {
    std::cout << "Cannot seek there:" << (long) pos << std::endl;
  }
  else
  {
    std::cout << "Seek to:" << (long) pos << std::endl;
  }
  return pos;
}

void CALLING_CONVENTION I_user_input_mxf_stream_task::updated_data(const uint8_t* data, size_t data_size)
{
  size_t nb_written = fwrite(data, 1, data_size, file); 
  if (nb_written != data_size)
  {
    std::cout << "Error upon writing of data" << std::endl;
  }
}

// A new header metadata was found in the file.
// Previous version of mxf_file is obsolete.
// You will be ensured of the completeness of the 
// I_mxf_file structure only when closed and complete will be set
// to true.
bool CALLING_CONVENTION I_user_input_mxf_stream_task::updated_mxf_file(I_mxf_file* pi_mxf_file, bool closed, bool complete, uint64_t)
{
  if (closed && complete)
  {
    return true;
  }
  else
  {
    return false;
  }
}

// **************************************************************************
// **************************************************************************

int main(int argc, char* argv[])
{
  if (argc < 2)
  {
    wcout << L"usage : in.mxf" << endl;
    exit(0);
  }

  I_mxf_file* pi_mxf_file = NULL;

  try
  {
    wcout << L"Retrieve MxfTk Error Handler" << endl;
    if (!GetMxfErrorHandlerInterface(&err_h))
      throw 1;

    wcout << L"MxfTk Initialization" << endl;
    InitMXF_TK();
    throw_on_error();

    wcout << L"Init mxf streaming device with " << argv[1] << endl;
    I_user_input_mxf_stream_task in_mxf(argv[1]);

    // Read the stream and stop as soon as we get a closed and complete header metadata.
    NewMxfStreamInterface(&pi_mxf_file, &in_mxf);
    throw_on_error();

    // for demonstration purpose we remove all the existing metadata
    // tracks and replace them with a simple static one.
    // However MXFTk let's you add and remove not only the whole trakcs
    // but also add/remove metadata_material and metadata sets.
    wcout << L"Removing previous metadata tracks" << endl;
    // loop through output material
    generic_list* genl = pi_mxf_file->output_material();
    I_generic_material** p_gen_mat = NULL;
    while ((p_gen_mat = genl->next()) != NULL)
    {
      I_generic_material* gen_mat = *p_gen_mat;
      // loop through metadata tracks
      track_list* trl = gen_mat->metadata_tracks();
      if (trl)
      {
        I_track** p_tr = NULL;
        while ((p_tr = trl->next()) != NULL)
        {
          I_track* tr = *p_tr;
          // remove metadata track
          gen_mat->delete_metadata_track(tr);
        }
        gen_mat->free_tracks(trl);
      }
    }
    throw_on_error();

    // add a simple metadata track to the first material
    genl->begin();
    I_generic_material* first_mat = *genl->next();
    pi_mxf_file->free_generic_list(genl);
    rational* pi_rat = NULL;
    NewRationalInterface(&pi_rat, 25, 1);
    I_track* new_track = first_mat->new_metadata_track(timeline_metadata, 0, pi_rat);
    FreeRationalInterface(&pi_rat);
    throw_on_error();

    // Create a Production framework
    I_metadata* pr_fr = NULL;
    NewMetadataInterfaceByLabel(&pr_fr, "http://www.smpte-ra.org/schemes/434/200X/groups/S380M/2004:ProductionFramework");
    throw_on_error();
    // add a simple property
    wstring wstr = L"The file was updated!";
    ADD_PROPERTY(pr_fr, "http://www.smpte-ra.org/schemes/434/200X/properties/RP210:FrameworkTitle", MXF_UTF16_STRING, wstr.size()*sizeof(wchar_t), wstr.c_str());

    // create the metadata material
    I_metadata_material* pi_meta_mat = NULL;
    NewMetadataMaterialInterface(&pi_meta_mat, // metata material
      m_timeline, // goes on a timeline track
      pr_fr, // production framework
      0, // origin
      -1, //duration will be automatically computed this way
      NULL, // no comment
      0); // size of comment = 0

    // add the metadata material to the metadata track.
    new_track->add_metadata(pi_meta_mat, NULL);

    UpdateAndFreeMxfFileInterface(&pi_mxf_file);
    throw_on_error();
  }
  catch(int)
  {
    wcout << L"An error occured!" << endl;
    // Retrieving details of the error(s) and display it(them).
    display_errors();
  }

  return 0;
}

void throw_on_error()
{
  if (err_h->has_an_error())
    throw 1;
}

void display_errors()
{
  // Retrieve error handler
  if (!err_h)
    return;

  // Retrieve list of errors
  error_list* err_l = err_h->get_errors();
  if (err_l)
  {
    err_l->begin();
    I_mxf_error** ppi_mxf_error = NULL;
    while ((ppi_mxf_error = err_l->next()) != NULL) //go through the list
    {
      I_mxf_error* pi_error = *ppi_mxf_error;
      // and display each error 
      wcout << L"Error " << (int)pi_error->get_id() << L": " << pi_error->get_error_msg() << endl;
    }
    // free error list interface
    err_h->free_errors(err_l);
  }
}
