//------------------------------------------------------------------------------
//  Author:    	Thierry FRANZETTI
//  Comment:	MXF_TK: An example showing how to add metadata from an XML file
//  to a newly generated MXF file.
//
// Copyright (c) 2008 OPENCUBE. All rights reserved.
// -----------------------------------------------------------------------------
#if defined(BC6)	// Borland C Builder 6.0
#include <vcl.h>
#endif

#include <mxf_tk/I_concrete_material.hpp>
#include <mxf_tk/I_metadata_material.hpp>
#include <mxf_tk/I_op1a_file.hpp>

#include <iostream>

using namespace opencube::MXFTK_NAMESPACE;
using namespace std;

// MxfTk error handler
I_mxf_error_handler* err_h = NULL;

//Throws an exception when an error occured 
void throw_on_error(); 

//Called if an error occured.
void display_errors();

// Function that will add metadata from an xml file to the mxf file.
void add_metadata(I_op1a_file* pi_op1a_file, const char* xml_file);

int main(int argc, char* argv[])
{
  if (argc < 5)
  {
    wcout << L"usage : out.mxf out.xml metadata_in.xml [source_files]" << endl;
    exit(0);
  }

  // the op1a mxf file we are going to write.
  I_op1a_file* pi_op1a_file = NULL;
  // the interface to manipulate the source to be wrapped.
  I_concrete_material* pi_conc_mat = NULL;
  // this boolean is set to true if pi_conc_mat needs to be destroyed
  bool delete_conc_mat = false;
  // Path to the source files to be wrapped
  locators* pi_loc = NULL;

  try
  {
    wcout << L"Retrieve MxfTk Error Handler" << endl;
    if (!GetMxfErrorHandlerInterface(&err_h))
      throw 1;

    wcout << L"MxfTk Initialization" << endl;
    InitMXF_TK();
    throw_on_error();

    wcout << L"Create the op1a mxf file interface" << endl;
    NewOp1aFileInterface(&pi_op1a_file, argv[1]);
    throw_on_error();

    wcout << L"Build source files list" << endl;
    GetEmptyLocatorInterface(&pi_loc);
    throw_on_error();
    for (int i=4; i<argc; i++)
      pi_loc->add((const char**)&(argv[i]));

    wcout << L"Create source material" << endl;
    NewConcreteMaterialInterface(&pi_conc_mat, pi_loc, frame_wrapping);
    throw_on_error();
    delete_conc_mat = true;

    wcout << L"Attach material to the file" << endl;
    pi_op1a_file->set_material(pi_conc_mat); // pi_conc_mat will be deleted upon destruction of pi_op1a_file
    throw_on_error();
    delete_conc_mat = false;

    // Metadata track will have the same origin and edit rate than the output material's tracks
    // retrieve these values
    wcout << L"Add XML metadata to the file" << endl;
    add_metadata(pi_op1a_file, argv[3]);
    throw_on_error();

    wcout << L"Write on disk" << endl;
    pi_op1a_file->flush();
    throw_on_error();

    wcout << L"Wait termination" << endl;
    pi_op1a_file->wait_end_flush_thread();
    throw_on_error();

    wcout << L"Output header metadata XML file" << endl; // optional!
    pi_op1a_file->output_xml(argv[2]);
    throw_on_error();
  }
  catch (int)
  {
    wcout << L"An error occured!" << endl;
    // Retrieving details of the error(s) and display it(them).
    display_errors();
  }

  wcout << L"Close Process" << endl;
  if (pi_op1a_file)
    FreeOp1aFileInterface(&pi_op1a_file);
  if (pi_conc_mat && delete_conc_mat)
    FreeConcreteMaterialInterface(&pi_conc_mat);
  if (pi_loc)
    FreeLocatorInterface(&pi_loc);

  return 0;
}

void add_metadata(I_op1a_file* pi_op1a_file, const char* xml_file)
{
  // Output material
  I_generic_material* pi_gen_mat = NULL;

  // List of tracks from the output material
  track_list* gen_tracks = NULL;

  // Metadata material that will be added to the metadata track.
  I_metadata_material* pi_meta_material = NULL;
  bool delete_meta_material = false;  

  // Root of XML metadata tree
  I_metadata* pi_framework = NULL;
  bool delete_framework = false;

  try
  {
    wcout << L"Get output material (metadata will be added to this material)" << endl;
    generic_list* gen_list = pi_op1a_file->output_material();
    gen_list->begin();
    pi_gen_mat = *gen_list->next();
    pi_op1a_file->free_generic_list(gen_list);
    throw_on_error();

    wcout << L"Retrieve the list of tracks from the output material" << endl;
    gen_tracks = pi_gen_mat->source_tracks();
    throw_on_error();

    gen_tracks->begin();
    wcout << L"Retrieve first track from the list" << endl;
    I_track* gen_tr = *gen_tracks->next();
    if (gen_tr)
    {
      wcout << L"\tRetrieve origin and edit rate of the track" << endl;
      int64_t origin = gen_tr->origin();
      rational *edit_rate = gen_tr->edit_rate();
      throw_on_error();

      wcout << L"\tCreate new static metadata track with adequate origin and edit rate" << endl;
      I_track* static_tr = pi_gen_mat->new_metadata_track(static_metadata, origin, edit_rate);
      FreeRationalInterface(&edit_rate);
      throw_on_error();

      if (static_tr)
      {
        wcout << L"\t\tGenerate descriptive metadata from XML file" << endl;
        NewMetadataInterfaceByXML(&pi_framework, xml_file);
        throw_on_error();
        delete_framework = true;

        wcout << L"\t\tAttach metadata to a metadata material" << endl; 
        NewMetadataMaterialInterface(&pi_meta_material, m_static, pi_framework, 0, 0, NULL, 0);
        throw_on_error();
        delete_framework = false; // pi_framework will be deleted upon desctuction of pi_meta_material
        delete_meta_material = true;

        wcout << L"\t\tAttach metadata material to the static track" << endl;
        static_tr->add_metadata(pi_meta_material, gen_tracks /* tracks that this metadata material is documenting */);
        delete_meta_material = false; // pi_meta_material will be destroyed upon deletion of static_tr.
      }
      throw_on_error();
    }
  }
  catch (int)
  {
    // do nothing will be rethrown in the following throw_on_error() 
    // and catched in main()
  }

  if (delete_framework)
    FreeMetadataInterface(&pi_framework);

  if (delete_meta_material)
    FreeMetadataMaterialInterface(&pi_meta_material);

  if (pi_gen_mat && gen_tracks)
    pi_gen_mat->free_tracks(gen_tracks);

  throw_on_error();
}

void throw_on_error()
{
  if (err_h->has_an_error())
    throw 1;
}

void display_errors()
{
  // Retrieve error handler
  if (!err_h)
    return;

  // Retrieve list of errors
  error_list* err_l = err_h->get_errors();
  if (err_l)
  {
    err_l->begin();
    I_mxf_error** ppi_mxf_error = NULL;
    while ((ppi_mxf_error = err_l->next()) != NULL) //go through the list
    {
      I_mxf_error* pi_error = *ppi_mxf_error;
      // and display each error 
      wcout << L"Error " << (int)pi_error->get_id() << L": " << pi_error->get_error_msg() << endl;
    }
    // free error list interface
    err_h->free_errors(err_l);
  }
}
