//------------------------------------------------------------------------------
//  Author:    	MXF Team
//  Comment:	MXF_TK: This class defines the essence type interface
//              describing the essence of a concrete track.
//
//  The following convention is used for defining parameters and retrun values:
//      in : input parameter (will be used for computation in the function)
//      out : output parameter (will be modified by the function call)
//      in-out : input parameter that may be modified by the function call.
//      free : MXFTk user is reponsible for the deletion of this parameter
//             unless it is passed as a parameter of another function without 
//             the "free" tag.
//
// Copyright (c) 2008 OPENCUBE. All rights reserved.
// -----------------------------------------------------------------------------
#ifndef OC_I_ESSENCE_TYPE_HPP
#define OC_I_ESSENCE_TYPE_HPP

#include <mxf_tk/mxf_tk.hpp>

namespace opencube
{
  namespace MXFTK_NAMESPACE
  {
    enum essence_source { ess_unknown,                // unrecognized essence             
			  ess_d10,                    // d10
			  ess_d11,                    // d11
			  ess_dv_unknown,             // dv other than iec ans smpte
			  ess_dv_iec,                 // dv iec
			  ess_dv_cam_iec,             // dv iec from dvcam-1
			  ess_dv_smpte,               // dv smpte
			  ess_mpeg_es,                // mpeg elementary stream
			  ess_mpeg_pes,               // mpeg packetized elementary stream
			  ess_mpeg_ps,                // mpeg program stream 
			  ess_mpeg_ts,                // mpeg transport stream
			  ess_mpeg4,                  // mpeg4
			  ess_uncompressed_unknown,   // uncompressed other than sd, hd
			  ess_uncompressed_sd,        // uncompressed sd
			  ess_uncompressed_hd1080,    // uncompressed hd 1080 lines
			  ess_uncompressed_hd720,     // uncompressed hd 720 lines
			  ess_bwf,                    // broadcast wave
			  ess_aes3,                   // audio AES3
			  ess_a_law,                  // A-Law audio
                          ess_aiff,                   // AIFF audio
			  ess_jpeg2k,                 // Jpeg 2000 pictures
			  ess_vc3,                    // VC-3 / DNxHD
			  ess_vbi,                    // VBI
			  ess_anc};                   // ANC Data

    enum essence_format { for_unknown,      // unknown or not applicable
			  for_525_5994p,    // 59.94Hz NTSC progressive
			  for_525_5994i,    // 59.94Hz NTSC interlaced 
			  for_525_60i,      // 60Hz NTSC interlaced
			  for_625_50i,      // 50Hz PAL interlaced
			  for_625_50p,      // 50Hz PAL progressive
			  for_720_2398p,    // 23.98Hz HD 720 progressive
			  for_720_24p,      // 24Hz HD720 progressive
			  for_720_25p,      // 25Hz HD720 progressive
			  for_720_2997p,    // 29.97Hz HD720 progressive
			  for_720_30p,      // 30Hz HD720 progressive
			  for_720_50p,      // 50Hz HD720 progressive
			  for_720_5994p,    // 59.94Hz HD720 progressive
			  for_720_60p,      // 60Hz HD720 progressive
			  for_1080_2398p,   // 23.98Hz HD1080 progressive
			  for_1080_2398sf,  // 23.98Hz HD1080 segmented frame
			  for_1080_24p,     // 24Hz HD1080 progressive
			  for_1080_24sf,    // 24Hz HD1080 segmented frame
			  for_1080_25p,     // 25Hz HD1080 progressive
			  for_1080_25sf,    // 25Hz HD1080 segmented frame
			  for_1080_2997p,   // 29.97Hz HD1080 progressive
			  for_1080_2997sf,  // 29.97Hz HD1080 segmented frame
			  for_1080_30p,     // 30Hz HD1080 progressive
			  for_1080_30sf,    // 30Hz HD1080 segmented frame
			  for_1080_50i,     // 50Hz HD1080 interlaced
			  for_1080_50p,     // 50Hz HD1080 progressive
			  for_1080_5994p,   // 59.94Hz HD1080 progressive
			  for_1080_5994i,   // 59.94Hz HD1080 interlaced
			  for_1080_60i,     // 60Hz HD1080 interlaced
			  for_1080_60p };   // 60Hz HD1080 progressive

    enum sample_structure { samp_unknown,       // unknown or not applicable to this format
			    samp_4_1_1,         // 4:1:1
			    samp_4_2_0,         // 4:2:0
			    samp_4_2_2,         // 4:2:2
			    samp_4_4_4_4 };     // 4:4:4:4

    enum layout { layout_full_frame = 0,        
		  layout_separate_fields,
		  layout_single_field,
		  layout_mixed_fields,
		  layout_segmented_frame,
		  layout_unknown };

    enum electro_spatial_form { formulation_two_channel_default                   = 0,
				formulation_two_channel                           = 1,
				formulation_single_channel                        = 2,
				formulation_primary_secondary                     = 3,
				formulation_stereophonic                          = 4,
				formulation_single_channel_double_frequency       = 7,
				formulation_stereo_left_channel_double_frequency  = 8,
				formulation_stereo_right_channel_double_frequency = 9,
				formulation_multi_channel_default                 = 15,
				formulation_unknown                               = 100 };



    /* I_essence_type:
     * A class describing the essence embedded on a concrete track.
     * You can use it to identify the essence wrapped in your MXF file.
     */
    class I_essence_type
    {
    public:  
      /*
       * Return the type of the essence
       */
      virtual essence_source CALLING_CONVENTION get_type() const=0;

      // ******************************************************************
      // PICTURE TRACK ONLY

      /*
       * Return the video format
       */
      virtual essence_format CALLING_CONVENTION get_format() const=0;

      /*
       * Return bit rate per second (Mbps) (if applicable)
       * Returns 0 if bit rate is not defined for this kind of essence
       */
      virtual uint32_t CALLING_CONVENTION bit_rate() const=0;

      /*
       * Return video sample structure (if applicable)
       */
      virtual sample_structure CALLING_CONVENTION sampling() const=0;

      /*
       * Return aspect ratio (if applicable).
       * If unknown or not applicable returns NULL.
       */
      virtual rational* CALLING_CONVENTION aspect_ratio() const=0;

      /*
       * Return component depth (eg 8, 10 or 16)
       * Return 0 if not applicable.
       */
      virtual uint32_t CALLING_CONVENTION component_depth() const=0;

      /*
       * Return stored width
       * Return 0 if not applicable.
       */
      virtual uint32_t CALLING_CONVENTION stored_width() const=0;

      /*
       * Return stored height
       * Return 0 if not applicable.
       */
      virtual uint32_t CALLING_CONVENTION stored_height() const=0;

      /*
       * Return display width
       * Return 0 if not applicable.
       */
      virtual uint32_t CALLING_CONVENTION display_width() const=0;

      /*
       * Return display height
       * Return 0 if not applicable.
       */
      virtual uint32_t CALLING_CONVENTION display_height() const=0;

      /*
       * Return frame layout.
       */
      virtual layout CALLING_CONVENTION frame_layout() const=0;

      // ****************************************************************
      // AUDIO TRACK ONLY

      /*
       * Return audio sampling rate if present
       */
      virtual rational* CALLING_CONVENTION sampling_rate() const=0;

      /*
       * Return electro-spatial fomulation
       */
      virtual electro_spatial_form CALLING_CONVENTION electro_spatial_formulation() const=0;

      /*
       * Return number of channels
       */
      virtual uint32_t CALLING_CONVENTION nb_channels() const=0;
      
      /*
       * Return number of quantization bits
       */
      virtual uint32_t CALLING_CONVENTION quantization() const=0;

      /*
       * Return locators if essence is externally referenced.
       * The boolean is set to true if the locators are network locators (a path)
       * Note that external references may be text locators.
       * Returned pointer should not be freed.
       */
      virtual locators* CALLING_CONVENTION external_references(bool& is_network_loc /* out */) const=0;

      /*
       * Update the external references for the track described by this essence type.
       * Use this function to update the existing external references when doing a
       * partial restore or moving the external references.
       * Note that previous external references are removed from the file.
       * Also note that you are responsible for freeing the locators and you may do so
       * directly after calling this function.
       * This function will not operate if the file does not contain external references.
       * Return true if successful.
       */
      virtual bool CALLING_CONVENTION update_external_references(locators* /* in free */, 
                                                                 bool is_network_loc /* in */)=0;

      // ****************************************************************
    };
  }
}

#endif
