//------------------------------------------------------------------------------
//  Author:    	MXF Team
//  Comment:	MXF_TK: This class defines the generic material interface.
//
//  The following convention is used for defining parameters and retrun values:
//      in : input parameter (will be used for computation in the function)
//      out : output parameter (will be modified by the function call)
//      in-out : input parameter that may be modified by the function call.
//      free : MXFTk user is reponsible for the deletion of this parameter
//             unless it is passed as a parameter of another function without 
//             the "free" tag.
//
// Copyright (c) 2008 OPENCUBE. All rights reserved.
// -----------------------------------------------------------------------------
#ifndef OC_I_GENERIC_MATERIAL_HPP
#define OC_I_GENERIC_MATERIAL_HPP

#include <mxf_tk/mxf_tk.hpp>
#include <mxf_tk/streaming.hpp>
#include <mxf_tk/I_umid.hpp>

namespace opencube
{
  namespace MXFTK_NAMESPACE
  {
    /*
     * Defines the possible types of track
     * picture:				video only essence track
     * sound:				audio only essence track
     * data:				data essence track (may represent a track containing both audio and video)
     * timeline_metadata:	descriptive metadata (a timeline metadata defines a track where metadata
     is continuously appended - at any point in the track, metadata MUST be found).
     The metadata is always ordered in a time-linear fashion and metadata segments
     cannot overlap.
     * event_metadata:		descriptive metadata (an event metadata defines a track where metadata is
     not continuously appended - there might be metadata segments overlapping on 
     this track and some sections of the track may not contain metadata at all).
     * static_metadata:		descriptive metadata (without any time reference - origin and duration are 
     irrelevant for this track). It is used to set metadata that is global to the
     entire material.
    */
    enum track_type {picture, sound, data, timeline_metadata, event_metadata, static_metadata, track_type_error};

    class I_timecode_material;
	
    /* I_generic_material:
     * A class describing a generic material. 
     * If not derivated, it usually represents the output of an mxf_file.
     * In mxf terms: Material Package.
     */
    class I_generic_material
    {
    public:
      /*
       * Returns a list of tracks defining the essence data of this 
       * material. Returned tracks will be either picture, sound or data track_type.
       * They describe the output essence as seen/heard while "playing" the 
       * mxf_file. Returned I_track*s shouldn't be destroyed by the API user.
       * Note: In a concrete_material returned I_track*s can be statically casted into
       * I_concrete_track*s.
       * Returns NULL if an error occured.
       */
      virtual track_list* CALLING_CONVENTION source_tracks()=0;

      /*
       * Returns a list of tracks defining the descriptive metadata of this 
       * material. Returned tracks will be either timeline_metadata,
       * event_metadata or static_metadata track_type . They describe the output 
       * of the metadata essence as seen while "playing" the mxf_file. 
       * Returned I_track*s may be edited.
       * Returns NULL if an error occured.
       */
      virtual track_list* CALLING_CONVENTION metadata_tracks()=0;

      /*
       * Frees a list of tracks returned by source_tracks or metadata_tracks.
       * Returns true if successful
       */
      virtual bool CALLING_CONVENTION free_tracks(track_list* /* in */)=0;

      /*
       * Returns a timecode material interface defining the timecode of 
       * the file playout.
       * Returned pointer should not be deleted.
       * Returns NULL if an error occured.
       */
      virtual I_timecode_material* CALLING_CONVENTION get_timecode() const=0;

      /*
       * Sets a new timecode material. According to MXF format
       * specifications, the file playout must be continuous
       * (by extent, the timecode material passed should describe
       * a continuous timecode). However the timecode of a I_concrete_material
       * does not require to be continuous.
       * Returns false if an error occured
       */
      virtual bool CALLING_CONVENTION set_timecode(I_timecode_material* /* in */)=0;
		
      /*
       * Creates a new metadata track in this material and returns it. 
       * Metadata can be added to the returned track using the
       * function add_metadata from I_track interface.
       * track_type can be timeline_metadata, event_metadata or static_metadata
       * int64_t sets the origin of the track measured in edit units
       * rational sets the edit rate of the track (number of edit units in a second).
       * The rational must be freed by the API user after calling this function.
       * Returned track will be destroyed upon deletion of the material.
       * Returns NULL if an error occured
       */
      virtual I_track* CALLING_CONVENTION new_metadata_track(track_type /* in */, 
                                                             int64_t /* in */, 
                                                             rational* /* in free */)=0;
		
      /*
       * Similar to previous function but enables metadata creation on-the-fly when streaming.
       * Returns NULL if an error occured.
       */
      virtual I_track* CALLING_CONVENTION new_stream_metadata_track(I_input_metadata_stream_task* /* in free */, 
                                                                    track_type /* in */, 
                                                                    int64_t /* in */, 
                                                                    rational* /* in free */)=0;

      /*
       * Removes a metadata track from the file. Inoperative on picture, sound and data tracks.
       */
      virtual bool CALLING_CONVENTION delete_metadata_track(I_track* /* in */) =0;

      /*
       * Returns the material class name as a const char* ("Generic Material",
       * "Concrete Material", "Timecode Material" or "Metadata Material").
       * Returns NULL if an error occured.
       */
      virtual const char* CALLING_CONVENTION type() const=0;

      /*
       * Sets the name of this material that will be embedded in the file (utf16 format)
       * the size_type parameter is used to specify the size in bytes of this string.
       * API user can free the wchar_t pointer after calling this function.
       * Returns true if successful
       */
      virtual bool CALLING_CONVENTION set_material_name(const wchar_t* /* in free */, 
                                                        size_t /* in */)=0;

      /*
       * Gets the name of this track (utf16 format) and retireves its size.
       * The string returned will be destroyed upon deletion of the I_track*
       * Returns NULL if an error occured.
       */
      virtual const wchar_t* CALLING_CONVENTION get_material_name(size_t& /* out */) const=0;

      /*
       * Sets umid of this material.
       * It is the API user responsability to free the I_umid interface
       * using FreeUmidInterface function.
       * Returns true if successful.
       */
      virtual bool CALLING_CONVENTION set_umid(const I_umid* /* in free */)=0;

      /*
       * Returns the umid of the essence
       * It is the API user responsability to free the returned I_umid interface
       * using FreeUmidInterface function.
       * Returns NULL if an error occured.
       */
      virtual I_umid* /* free */ CALLING_CONVENTION get_umid()=0;
    };
  }
}

#endif
