//------------------------------------------------------------------------------
//  Author:    	Thierry FRANZETTI
//  Comment:	MXF_TK: This example illustrates mxf file creation while streaming
//				the input essences in active mode.
//              This is configured for creating XDCAM IMX files.
//
// Copyright (c) 2008 OPENCUBE. All rights reserved.
// -----------------------------------------------------------------------------
#if defined(BC6)	// Borland C Builder 6.0
#include <vcl.h>
#endif

#include <mxf_tk/I_concrete_material.hpp>
#include <mxf_tk/I_xdcam_imx_file.hpp>

#include <iostream>
#include <sstream>

#ifdef WIN32
#include <windows.h>
#else
#include <unistd.h>
#endif

// Size of the input stream buffer
#define BUF_SIZE 16384

using namespace opencube::MXFTK_NAMESPACE;
using namespace std;

// MxfTk error handler
I_mxf_error_handler* err_h = NULL;

//Throws an exception when an error occured 
void throw_on_error(); 

//Called if an error occured.
void display_errors();

/* Derivation of the pure vitual class from MxfKt
Note that the function data_request was not reimplemented.
See passive mode example for a demonstration of use of the function data_request.
Here we use an implementation using files for conveniency
however the same mechanism could be used with data originating
from a true streaming device (linear storage device, Firewire port) 
*/
class I_user_stream_task : public I_essence_stream_task
{
public:
  // Constructor -> init source streams
  I_user_stream_task(unsigned int);

  // Free source streams
  virtual ~I_user_stream_task();

  // Sets passive mode
  virtual bool CALLING_CONVENTION get_passive_mode() { return false; }

  // Sets the file information
  virtual void CALLING_CONVENTION get_stream_info(unsigned int stream_id, 
    essence_source& s, 
    bool& is_video, 
    unsigned int& nb_mpeg_es);

  // Never called in active mode ! 
  virtual size_t CALLING_CONVENTION data_request(unsigned int, uint8_t*, size_t) { return 0; }

  // This function will be called once by MxfTk to know how many source streams
  // are to be wrapped (for instance audio and video may originate from a different
  // Firewire port).
  virtual unsigned int CALLING_CONVENTION get_total_number_of_streams() { return _nb_streams; };

  // This function will be called once by MxfTk.
  // Sets the size of MxfTk internal buffer to bufferize input streams.
  // Here it is set to the input buffer size. Increasing its size may 
  // help to increase performances.
  // Set buffer size to 1Mb.
  virtual size_t CALLING_CONVENTION get_buffer_size(unsigned int) { return BUF_SIZE; };

protected:
  // Total number of essences (streams) to be wrapped.
  unsigned int _nb_streams;
};

// Initalize nb_streams streaming devices (files here) 
I_user_stream_task::I_user_stream_task(unsigned int nb_streams)
: I_essence_stream_task(), _nb_streams(nb_streams)
{
}

// Free streaming devices.
I_user_stream_task::~I_user_stream_task()
{
}

void CALLING_CONVENTION I_user_stream_task::get_stream_info(unsigned int stream_id, 
                                                            essence_source& s, 
                                                            bool& is_video,
                                                            unsigned int& nb_mpeg_es)
{ 
  if (stream_id == 0)
  {
    s = ess_d10;
    is_video = true;
    nb_mpeg_es = 0;
  }
  else if (stream_id == 1)
  {
    s = ess_d10;
    is_video = false;
    nb_mpeg_es = 0;
  }
}

//*************************************************************************************************************
//*************************************************************************************************************
// main loop

int main(int argc, char* argv[])
{
  if (argc != 4 )
  {
    cout << "usage : out.mxf imx_source file aes_source file" << endl;
    exit(0);
  }

  I_xdcam_imx_file* pi_xdcam_file = NULL;
  I_concrete_material* pi_conc_mat = NULL;

  try
  {
    wcout << L"Retrieve MxfTk Error Handler" << endl;
    if (!GetMxfErrorHandlerInterface(&err_h))
      throw 1;

    wcout << L"MxfTk Initialization" << endl;
    InitMXF_TK();
    throw_on_error();

    cout << "Create the Xdcam IMX file" << endl;
    NewXdcamImxFileInterface(&pi_xdcam_file, argv[1]);
    throw_on_error();

    wcout << L"Create input streaming device" << endl;
    // argc-2 is the number of input streams.
    I_user_stream_task s_task(argc-2);

    wcout << L"Create source material" << endl;
    // Frame wrapping is always recommended to create streamable files.
    NewStreamMaterialInterface(&pi_conc_mat, &s_task, xdcam_wrapping);
    throw_on_error();

    wcout << L"Attach material to the file" << endl;
    pi_xdcam_file->set_material(pi_conc_mat); 
    throw_on_error();

    // Create streaming buffers (from file)
    FILE** files = new FILE*[argc-2]; // list of files - handles to the streams
    uint8_t** buffers = new uint8_t*[argc-2]; // buffers from the streams
    size_t* buffers_size = new size_t[argc-2]; // amount of data stored in each buffer
    bool* previous_buffer_written = new bool[argc-2]; // did we write the buffer?
    bool* end_of_stream = new bool[argc-2]; // did we write all?
    for (unsigned int i=0; i< (unsigned int)(argc-2); i++)
    {
      files[i] = fopen(argv[i+2], "rb");	
      buffers[i] = new uint8_t[BUF_SIZE];
      buffers_size[i] = 0;
      previous_buffer_written[i] = true;
      end_of_stream[i] = false;
    }

    wcout << L"Launch the flush thread" << endl;
    pi_xdcam_file->flush(); 
    throw_on_error();

    // Send data as long as there is some to send
    // and wait for thread completion
    bool end = false;
    while (!end)
    {
      // boolean used to know if at least one buffer was written
      bool could_we_write = false;
      end = true;

      for (unsigned int i=0; i< (unsigned int)(argc-2); i++)
      {
        if (!end_of_stream[i]) // is there something left to write?
        {
          // yes so we haven't finished yet
          end = false;

          // do we have a stream buffer left to write ?
          if (previous_buffer_written[i])
          {
            // retrieve essence data (here it's read from a file for demonstration purpose
            // but in a real streaming environment this would buffers from your 
            // streaming device (Firewire, graphics card buffer, etc...)
            buffers_size[i] = fread(buffers[i], 1, BUF_SIZE, files[i]);
          }

          // is there some data left on the stream ?
          if (buffers_size[i]>0)
          {
            // yes then give it to MXFTk 
            size_t nb_written = pi_conc_mat->write(i, buffers[i], buffers_size[i]);
            // could we write the data
            if (nb_written==buffers_size[i])
            {
              buffers_size[i] = 0;
              // and remember we could write a buffer
              could_we_write = true;
              previous_buffer_written[i] = true;
            }
            else if (nb_written==0)
            {
              previous_buffer_written[i] = false;
            }
            else if (nb_written!=0) // we wrote only a part of the buffer -> should never occur
            {
              throw(1);
            }
          }
          else
          {
            // no then notifies the end of the stream to MXFTk
            pi_conc_mat->end_of_stream(i);
            could_we_write = true;
            buffers_size[i] = 0;
            end_of_stream[i] = true;
            fclose(files[i]);
          }
        }
      }

      // could we write a buffer
      if (!could_we_write)
      {
        throw_on_error();
        // check the thread is still running!
        if (pi_xdcam_file)
        {
          end = end || (!pi_xdcam_file->flush_thread_running());
        }
        // no -> so it probably means we are trying to write 
        // faster than MXFTk can wrap or we have nothing left to write
        // wait a little bit
#ifdef WIN32
        Sleep(5);
#else
	usleep(5000);
#endif
      }
    }

    // free streaming buffers
    for (unsigned int i=0; i< (unsigned int)(argc-2); i++)
    {
      delete [] buffers[i];
    }
    delete [] files;
    delete [] buffers;
    delete [] buffers_size;
    delete [] previous_buffer_written;
    delete [] end_of_stream;

    // check for errors
    throw_on_error();

    // wait for complete stop of the flush thread (you MUST call
    // this function to deallocate the memory allocated by the flush.
    pi_xdcam_file->wait_end_flush_thread();

    wcout << L"Close Process" << endl;
  }
  catch(int)
  {
    wcout << L"An error occured!" << endl;
    // Retrieving details of the error(s) and display it(them).
    display_errors();
  }

  if (pi_xdcam_file)
    FreeXdcamImxFileInterface(&pi_xdcam_file);

  return 0;
}

void throw_on_error()
{
  if (err_h->has_an_error())
    throw 1;
}

void display_errors()
{
  // Retrieve error handler
  if (!err_h)
    return;

  // Retrieve list of errors
  error_list* err_l = err_h->get_errors();
  if (err_l)
  {
    err_l->begin();
    I_mxf_error** ppi_mxf_error = NULL;
    while ((ppi_mxf_error = err_l->next()) != NULL) //go through the list
    {
      I_mxf_error* pi_error = *ppi_mxf_error;
      // and display each error 
      wcout << L"Error " << (int)pi_error->get_id() << L": " << pi_error->get_error_msg() << endl;
    }
    // free error list interface
    err_h->free_errors(err_l);
  }
}
