//------------------------------------------------------------------------------
//  Author:    	Thierry FRANZETTI
//  Comment:	MXF_TK: This example illustrates mxf file creation while streaming
//				the input essences in passive mode.
//              This is configured for creating DCI OpAtom files.
//
// Copyright (c) 2008 OPENCUBE. All rights reserved.
// -----------------------------------------------------------------------------
#if defined(BC6)	// Borland C Builder 6.0
#include <vcl.h>
#endif

#include <mxf_tk/I_concrete_material.hpp>
#include <mxf_tk/I_opatom_file.hpp>

#include <string>
#include <iostream>
#include <sstream>

#ifdef WIN32
#include <windows.h>
#endif

// Size of the input stream buffer
#define BUF_SIZE 16384

using namespace opencube::MXFTK_NAMESPACE;
using namespace std;

// MxfTk error handler
I_mxf_error_handler* err_h = NULL;

//Throws an exception when an error occured 
void throw_on_error(); 

//Called if an error occured.
void display_errors();

//*************************************************************************************************************
/* Derivation of the pure vitual class from MxfKt
Note that the function data_request was not reimplemented.
See passive mode example for a demonstration of use of the function data_request.
Here we use an implementation using files for conveniency
however the same mechanism could be used with data originating
from a true streaming device (linear storage device, Firewire port) 
*/
class I_crypted_user_stream_task : public I_crypted_essence_stream_task
{
public:
  // Constructor -> init source streams
  I_crypted_user_stream_task(char* source_file, bool crypt, 
			     const uint8_t * cryptographic_key_id,
			     const uint8_t * cipher_key);

  // Free source streams
  virtual ~I_crypted_user_stream_task();

  // Sets passive mode
  virtual bool CALLING_CONVENTION get_passive_mode() { return true; }

  // Sets the file information
  virtual void CALLING_CONVENTION get_stream_info(unsigned int stream_id, 
						  essence_source& s, 
						  bool& is_video, 
						  unsigned int& nb_mpeg_es);
  
  // Sets the MPEG PS/TS substream info
  virtual void CALLING_CONVENTION get_mpeg_es_info(unsigned int stream_id, 
						   unsigned int sub_mpeg_es_id, 
						   essence_source& s, 
						   bool& is_video);

  // Always called in passive mode ! -> you must implement it!
  virtual size_t CALLING_CONVENTION data_request(unsigned int, uint8_t*, size_t);

  // This function will be called once by MxfTk to know how many source streams
  // are to be wrapped (for instance audio and video may originate from a different
  // Firewire port).
  virtual unsigned int CALLING_CONVENTION get_total_number_of_streams() { return 1; };

  // This function will be called once by MxfTk.
  // Sets the size of MxfTk internal buffer to bufferize input streams.
  // Here it is set to the input buffer size. Increasing its size may 
  // help to increase performances.
  // Set buffer size to 1Mb.
  virtual size_t CALLING_CONVENTION get_buffer_size(unsigned int) { return BUF_SIZE; };

  //Returns if stream_id is crypted. If no, following functions won't be called 
  //for this stream id
  virtual bool CALLING_CONVENTION do_crypt(unsigned int/* stream_id*/) { return _do_crypt; }

  //Returns the plaintext offset
  //plainttext offset is the offset of the crypted data within a frame.
  //it is the size of the plaintext at start of each frame.
  virtual uint32_t CALLING_CONVENTION get_plaintext_offset(unsigned int/* stream_id*/) { return 0; }

  //Returns the cryptographic key id, this value is stored in mxf file
  //and is used by mxf decoder to retrieve cipher key.
  //an external key manager give the cipher key corresponding to a cryptographic key id
  virtual const uint8_t * CALLING_CONVENTION get_cryptographic_key_id(unsigned int/* stream_id*/) { return _crypt_id; }

  //Returns the cipher key.
  //Cipher key is used to crypt data but is not stored in mxf file.
  //Decoding is possible by giving the same cipher to the decoder.
  //Cipher key is retrieved from the cryptographic key id and is provided by the key manager.
  virtual const uint8_t * CALLING_CONVENTION get_cipher_key(unsigned int/* stream_id*/) { return _cipher; }

protected:
  // File handle (simulating true stream)
  FILE * file;
  bool _do_crypt;
  uint8_t * _crypt_id;
  uint8_t * _cipher;
};

// Initalize nb_streams streaming devices (files here) 
I_crypted_user_stream_task::I_crypted_user_stream_task(char* source_file, bool crypt, 
						       const uint8_t * cryptographic_key_id,
						       const uint8_t * cipher_key)
  : I_crypted_essence_stream_task(), _do_crypt(crypt), _crypt_id(NULL), _cipher(NULL)
{
  if(crypt)
  {
    if(cryptographic_key_id)
    {
      _crypt_id = new uint8_t[16];
      memcpy(_crypt_id, cryptographic_key_id, 16);
    }
    if(cipher_key)
    {
      _cipher = new uint8_t[16];
      memcpy(_cipher, cipher_key, 16);
    }
  }

  // Open streaming devices.
  file = fopen(source_file, "rb");
  if (ferror(file))
  {
    wcout << L"An error occured while opening the file " << source_file << endl;
  }
}

// Free streaming devices.
I_crypted_user_stream_task::~I_crypted_user_stream_task()
{
  if(_crypt_id)
  {
    delete[] _crypt_id;
  }
  if(_cipher)
  {
    delete[] _cipher;
  }

  fclose(file);
}

size_t CALLING_CONVENTION I_crypted_user_stream_task::data_request(unsigned int/* stream_id*/, 
								   uint8_t* buff, 
								   size_t s_req)
{
  try
  {
    // If the streaming device is not ready the current process can be put asleep
    // with a mutex and woke up only once some data is ready.

    // Retrieve data from streaming device buffer
    size_t s_read = fread(buff, sizeof(uint8_t), s_req, file);
    if (ferror(file)) 
    {
      throw 1;
    }

    return s_read;
  }
  catch(int)
  {
    wcout << L"An error occured!" << endl;
    // Retrieving details of the error(s) and display it(them).
    display_errors();
  }
  return 0;
}

void CALLING_CONVENTION I_crypted_user_stream_task::get_stream_info(unsigned int/* stream_id*/, 
								    essence_source& s, 
								    bool& is_video,
								    unsigned int& nb_mpeg_es)
{ 
  // TO DO : set different types
  s = ess_mpeg_es;
  is_video = true;
  nb_mpeg_es = 1;
}

void CALLING_CONVENTION I_crypted_user_stream_task::get_mpeg_es_info(unsigned int/* stream_id*/, 
								     unsigned int/* sub_mpeg_es_id*/, 
								     essence_source&/* s*/, 
								     bool& is_video)
{
  // TO DO : set different types
  is_video = true;
}

//*************************************************************************************************************
/* Derivation of the pure vitual class from MxfKt
Note that the function data_request was not reimplemented.
See passive mode example for a demonstration of use of the function data_request.
Here we use an implementation using files for conveniency
however the same mechanism could be used with data originating
from a true streaming device (linear storage device, Firewire port) 
*/

//*************************************************************************************************************
//*************************************************************************************************************
// main loop

int main(int argc, char* argv[])
{
  if (argc != 4)
  {
    cout << "usage : out.mxf mpeg2_source_file crypting" << endl;
    cout << "\tcrypting id \"true\" or \"false\"" << endl;
    exit(0);
  }

  string do_crypt_str(argv[3]);
  if(do_crypt_str != string("true") && do_crypt_str != string("false"))
  {
    cout << "usage : out.mxf mpeg2_source_file crypting" << endl;
    cout << "\tcrypting id \"true\" or \"false\"" << endl;
    exit(0);
  }

  bool do_crypt = (do_crypt_str == string("true"));

  uint8_t * cryptographic_key_id = new uint8_t[16];
  uint8_t * cipher_key = new uint8_t[16];
  
  for(int ii = 0; ii < 16; ii++)
  {
    cryptographic_key_id[ii] = ii;
    cipher_key[ii] = 15 - ii;
  }

  I_opatom_file* pi_opatom_file = NULL;
  I_concrete_material* pi_conc_mat = NULL;

  try
  {
    wcout << L"Retrieve MxfTk Error Handler" << endl;
    if (!GetMxfErrorHandlerInterface(&err_h))
    {
      throw 1;
    }

    wcout << L"MxfTk Initialization" << endl;
    InitMXF_TK();
    throw_on_error();

    cout << "Create the OpAtom file" << endl;
    NewDCPOpAtomFileInterface(&pi_opatom_file, argv[1]);
    throw_on_error();

    wcout << L"Create input streaming device" << endl;
    I_crypted_user_stream_task s_task(argv[2], do_crypt, cryptographic_key_id, cipher_key);

    wcout << L"Create source material" << endl;
    // Frame wrapping is required in DCI
    NewStreamMaterialInterface(&pi_conc_mat, &s_task, dcp_wrapping);
    throw_on_error();

    wcout << L"Attach material to the file" << endl;
    pi_opatom_file->set_material(pi_conc_mat); 
    throw_on_error();

    wcout << L"Launch the flush thread" << endl;
    pi_opatom_file->flush(); 
    throw_on_error();
    // wait for complete stop of the flush thread (you MUST call
    // this function to deallocate the memory allocated by the flush).
    pi_opatom_file->wait_end_flush_thread();
    wcout << L"Close Process" << endl;
  }
  catch(int)
  {
    wcout << L"An error occured!" << endl;
    // Retrieving details of the error(s) and display it(them).
    display_errors();
  }

  if (pi_opatom_file)
  {
    FreeOpAtomFileInterface(&pi_opatom_file);
  }

  delete[] cryptographic_key_id;
  delete[] cipher_key;

  return 0;
}

void throw_on_error()
{
  if (err_h->has_an_error())
  {
    throw 1;
  }
}

void display_errors()
{
  // Retrieve error handler
  if (!err_h)
  {
    return;
  }

  // Retrieve list of errors
  error_list* err_l = err_h->get_errors();
  if (err_l)
  {
    err_l->begin();
    I_mxf_error** ppi_mxf_error = NULL;
    while ((ppi_mxf_error = err_l->next()) != NULL) //go through the list
    {
      I_mxf_error* pi_error = *ppi_mxf_error;
      // and display each error 
      wcout << L"Error " << (int)pi_error->get_id() << L": " << pi_error->get_error_msg() << endl;
    }
    // free error list interface
    err_h->free_errors(err_l);
  }
}
