//------------------------------------------------------------------------------
//  Author:    	Thierry FRANZETTI
//  Comment:	MXF_TK: An example to display the content of an mxf file and 
//  extract the essence embedded.
//
// Copyright (c) 2008 OPENCUBE. All rights reserved.
// -----------------------------------------------------------------------------
#if defined(BC6)	// Borland C Builder 6.0
#include <vcl.h>
#endif

#include <mxf_tk/I_concrete_track.hpp>
#include <mxf_tk/I_timecode_material.hpp>
#include <mxf_tk/I_mxf_error.hpp>
#include <mxf_tk/I_mxf_file.hpp>

#include <iostream>
#include <sstream>
#include <fstream>

using namespace opencube::MXFTK_NAMESPACE;
using namespace std;

// MxfTk error handler
I_mxf_error_handler* err_h = NULL;

//Throws an exception when an error occured 
void throw_on_error(); 

//Called if an error occured.
void display_errors();

//Display MXF file content
void display_embedded_material(I_mxf_file*, concrete_list*, std::string);
/* essence extractor -> write audiovisual material on the disk */
void read_essence(I_mxf_file*, I_concrete_track*, std::string);

int main(int argc, char* argv[])
{
  if (argc != 3)
  {
    wcout << L"usage : out_base_name in.mxf" << endl;
    exit(0);
  }

  I_mxf_file* pi_mxf_file = NULL;

  try
  {
    wcout << "Retrieve MxfTk Error Handler" << endl;
    if (!GetMxfErrorHandlerInterface(&err_h))
      throw 1;

    wcout << L"MxfTk Initialization" << endl;
    InitMXF_TK();
    throw_on_error();
    // remove commentaries to set KLV by KLV access when
    // unwrapping custom wrapped MXF files (XDCAM proxy for
    // instance).
    //ExtractCustomByKLV(true);

    wcout << L"Loading " << argv[2] << endl;
    NewMxfFileInterface(&pi_mxf_file, argv[2]);
    throw_on_error();

    wcout << L"Operational Pattern: " << pi_mxf_file->operational_pattern() << endl;

    // check if it contains external references!
    if (pi_mxf_file->has_external_essence())
    {
      wcout << L"File contains external reference(s)." << endl;
    }

    concrete_list* embedded_mat = pi_mxf_file->embedded_material();
    wcout << L"Top-Level Source Packages (embedded material): " << (long)embedded_mat->size() << endl;
    display_embedded_material(pi_mxf_file, embedded_mat, std::string(argv[1]));
    pi_mxf_file->free_concrete_list(embedded_mat);
    throw_on_error();
  }
  catch(int)
  {
    wcout << L"An error occured!" << endl;
    // Retrieving details of the error(s) and display it(them).
    display_errors();
  }

  if (pi_mxf_file)
    FreeMxfFileInterface(&pi_mxf_file);

  return 0;
}                               

void display_embedded_material(I_mxf_file* pi_mxf_file, concrete_list* embedded_mat, std::string name)
{
  wcout << endl << L"Embedded Material :" << endl;
  // loop through embedded material.
  embedded_mat->begin();
  I_concrete_material** pi_mat_p = NULL;
  while ((pi_mat_p = embedded_mat->next())!=NULL)
  {
    I_concrete_material* pi_conc = *pi_mat_p;

    size_t s;
    const wchar_t* name_m = pi_conc->get_material_name(s);
    if (name_m)
      wcout << L"Package " << name_m << endl;
    else
      wcout << L"New Package " << endl;

    wcout << L"Timecode :" << endl;
    I_timecode_material* tm_mat = pi_conc->get_timecode();
    if (tm_mat)
    {
      rational *editrate = tm_mat->get_edit_rate();
      double ed = editrate->get_double();	   
      int64_t origin = tm_mat->get_origin();
      int64_t duration = tm_mat->get_total_duration();
      FreeRationalInterface(&editrate);

      wstringstream str_time;
      str_time << L"\tTimecode starts at " << (long) origin << L"eu lasts " << (long)duration << L"eu @" << (double)ed << L"eu/s" << endl;
      wcout << str_time.str();

      // Retrieve list of timecode components
      ordered_timecode_list* vect_time = tm_mat->get_timecode_list();
      vect_time->begin();
      I_timecode_component** time_p = NULL;
      while ((time_p = vect_time->next())!=NULL)
      {
        wstringstream tmstr;
        I_timecode_component* tmc = *time_p;
        tmstr << L"\t" << (tmc->get_timecode())->get_time() << L" lasts " << (long)tmc->get_duration() << L" eu" << endl;
        wcout << tmstr.str();
      }
      tm_mat->free_timecode_list(vect_time);
    }

    // display universal identifier UMID of this material
    I_umid* um = pi_conc->get_umid();
    wcout << L"\tumid " << um->get_str() << endl;
    FreeUmidInterface(&um);

    // loop through tracks
    wcout << L"\tTracks: " << endl;
    track_list* source_tr = pi_conc->source_tracks();
    source_tr->begin();
    I_track** trp = NULL;
    while ((trp = source_tr->next())!=NULL)
    {
      I_concrete_track* ctr = (I_concrete_track*)*trp;
      wstringstream str;

      rational *r = ctr->edit_rate();
      double rate = r->get_double();
      FreeRationalInterface(&r);
      str << L"\t\t[ " << ctr->type_name() << L" : start " << (long)ctr->origin() << L" lasts " << (long) ctr->duration() << L"eu @" << (double) rate << L"eu/s" << L" ]" << endl;
      wcout << str.str();

      // Extract essence in a file
      read_essence(pi_mxf_file, ctr, name);		
    }

    pi_conc->free_tracks(source_tr);
    wcout << endl;
  }
}

void read_essence(I_mxf_file* pi_mxf_file, I_concrete_track* tr, std::string name)
{
  wcout << L"\t\tRetrieve information on the track's content" << endl;
  I_essence_type* ess_type = tr->get_essence_type();
  throw_on_error();
  if (!ess_type)
    return;

  // set to true if series of images
  bool multifiles(false);

  // Retrieve essence type to write with correct extension
  std::string file_extension = "unknown";
  switch (ess_type->get_type())
  {     
  case ess_d10:
    {
      if (tr->type()==sound)
      {
        wcout << L"\t\tD10 Audio AES3"; 
        file_extension = ".aes";
        break;
      }
      else if (tr->type()==picture)
      {
        wcout << L"\t\tD10 Video IMX";
        file_extension = ".imx";
        break;
      }
    }
  case ess_dv_iec:                 
  case ess_dv_cam_iec:             
    {
      wcout << L"\t\tDV IEC";
      file_extension = ".dv";
      break;
    }
  case ess_dv_smpte:               
    {
      wcout << L"\t\tDV SMPTE";
      file_extension = ".dif";
      break;
    }
  case ess_mpeg_es:
    {
      wcout << L"\t\tMPEG Elementary Stream";
      file_extension = ".m2v";
      break;
    }
  case ess_mpeg_pes:
    {
      wcout << L"\t\tMPEG Packetized Elementary Stream";
      file_extension = ".mpg";
      break;
    }
  case ess_mpeg_ps:
    {
      wcout << L"\t\tMPEG Program Stream";
      file_extension = ".mpg";
      break;
    }
  case ess_mpeg_ts:                  
    {
      wcout << L"\t\tMPEG Transport Stream";
      file_extension = ".mpg";
      break;
    }
  case ess_mpeg4:
    {
      wcout << L"\t\tMPEG4 Stream";
      file_extension = ".m4v";
      break;
    }
  case ess_bwf:
    {
      wcout << L"\t\tBroadcast Wave";
      file_extension = ".wav";
      break;
    }
  case ess_aiff:
    {
      wcout << L"\t\tAiff";
      file_extension = ".aiff";
      break;
    }
  case ess_aes3: 
    {
      wcout << L"\t\tAudio AES3";
      file_extension = ".aes";
      break;
    }
  case ess_a_law: 
    {
      wcout << L"\t\tAudio A-LAW";
      file_extension = ".wav";
      break;
    }
  case ess_uncompressed_unknown:   
  case ess_uncompressed_sd:
  case ess_uncompressed_hd1080:
  case ess_uncompressed_hd720: 
    {
      wcout << L"\t\tSeries of Images";
      file_extension = ".up"; // correct file extension cannot be determined
      multifiles = true;
      break;
    }
  case ess_jpeg2k:
    {
      wcout << L"\t\tJPEG2000";
      file_extension = ".j2k";
      multifiles = true;
      break;
    }
  case ess_vc3:
    {
      wcout << L"\t\tVC-3";
      file_extension = ".hd";
      multifiles = true;
      break;
    }
  default:
    {
      wcout << L"\t\tUnknown" << endl;
      break;
    }
  }

  // check if the essence is external!
  if (pi_mxf_file->has_external_essence())
  {
    bool network_loc(false);
    locators* loc = ess_type->external_references(network_loc);
    if (loc && loc->size()>0)
    {
      loc->begin();
      if (network_loc)
      {
        wcout << L" (External Netwotk Locator)" << std::endl;
        const char** pextfile = loc->next();
        // Attach ext file to the track
        if (pextfile)
        {
          const char* extfile = *pextfile;
          tr->set_external_ref_file(extfile);
          throw_on_error();
        }
      }
      else
      {
        wcout << L" (External Text Locator)" << std::endl;
      }
    }
  }
  else
  {
    wcout << std::endl;
  }

  stringstream s_id;
  s_id << (long)tr->get_track_id();
  string str_id = "_" + s_id.str();
  if (!multifiles)
  {
    filebuf ofs;
    ofs.open( (name + str_id + file_extension).c_str(), ios::out | ios::binary  );

    // get track's origin
    int64_t ori = tr->origin();

    wcout << L"\t\tSeek at the beginning of the track" << endl;
    tr->seek_eu(-ori);

    // loop as long as some data is ready to be read.
    wcout << L"\t\tWriting essence file edit unit by edit unit" << endl;
    uint64_t eu_size= 0;
    while ((eu_size = tr->current_eu_size())!=0)
    {
      uint8_t* buffer = new uint8_t[size_t(eu_size)];
      size_t read = tr->read_eu(buffer, size_t(eu_size));
      ofs.sputn((char*) &buffer[0], std::streamsize(read) );
      delete [] buffer;
      tr->seek_next_eu();
      throw_on_error();
    }

    ofs.close();
  }
  else
  { 
    // get track's origin
    int64_t ori = tr->origin();

    wcout << L"\t\tSeek at the beginning of the track" << endl;
    tr->seek_eu(-ori);

    // loop as long as some data is ready to be read.
    wcout << L"\t\tWriting each file edit unit by edit unit" << endl;
    uint64_t eu_size= 0;
    while ((eu_size = tr->current_eu_size())!=0)
    {
      uint8_t* buffer = new uint8_t[size_t(eu_size)];
      size_t read = tr->read_eu(buffer, size_t(eu_size));

      // build file name and write it!
      stringstream s_eu;
      s_eu << (long)tr->get_current_eu();
      string str_eu = "_" + s_eu.str();
      filebuf ofs;
      ofs.open( (name + str_id + str_eu + file_extension).c_str(), ios::out | ios::binary  );
      ofs.sputn((char*) &buffer[0], std::streamsize(read) );
      ofs.close();

      delete [] buffer;
      tr->seek_next_eu();
      throw_on_error();
    }
  }
}

void throw_on_error()
{
  if (err_h->has_an_error())
    throw 1;
}

void display_errors()
{
  // Retrieve error handler
  if (!err_h)
    return;

  // Retrieve list of errors
  error_list* err_l = err_h->get_errors();
  if (err_l)
  {
    err_l->begin();
    I_mxf_error** ppi_mxf_error = NULL;
    while ((ppi_mxf_error = err_l->next()) != NULL) //go through the list
    {
      I_mxf_error* pi_error = *ppi_mxf_error;
      // and display each error 
      wcout << L"Error " << (int)pi_error->get_id() << L": " << pi_error->get_error_msg() << endl;
    }
    // free error list interface
    err_h->free_errors(err_l);
  }
}
