//------------------------------------------------------------------------------
//  Author:    	Thierry FRANZETTI
//  Comment:	MXF_TK: This example illustrates partial restore of MXF files
//              (including MXF files with external references)
//              This example does not handle partial restore of 
//              OpAtom MXF file.
//
// Copyright (c) 2008 OPENCUBE. All rights reserved.
// -----------------------------------------------------------------------------
#if defined(BC6)	// Borland C Builder 6.0
#include <vcl.h>
#endif

#include <mxf_tk/mxf_tk.hpp>
#include <mxf_tk/I_mxf_file.hpp>
#include <mxf_tk/I_op1b_file.hpp>
#include <mxf_tk/I_timecode.hpp>
#include <mxf_tk/I_concrete_material.hpp>
#include <mxf_tk/I_timecode_material.hpp>

#include <iostream>
#include <sstream>
#include <vector>

using namespace opencube::MXFTK_NAMESPACE;
using namespace std;

// MxfTk error handler
I_mxf_error_handler* err_h = NULL;

//Throws an exception when an error occured 
void throw_on_error(); 

//Called if an error occured.
void display_errors();

// retrieve TC IN from MXF file
I_timecode* get_tc_in(string file);

// retrieve list of external references
// returns true if it contains external ref to MXF files
bool get_external_references(I_mxf_file* pi_mxf_file, vector<string>& lst);

// partial restore one file.
bool partial_restore(const char* file, I_timecode* tc_in, I_timecode* tc_out, vector<string>* lst_files, vector<I_mxf_file*>* lst_mxf_files);

//Timecode redefinition
void redefine_timecode(I_op1b_file*, I_timecode*);

int main(int argc, char* argv[])
{
  if (argc < 6 )
  {
    wcout << L"usage : in.mxf tc_in(xx:xx:xx:xx) tc_out(xx:xx:xx:xx) frame_rate drop_frame(yes|no)" << endl;
    exit(0);
  }

  // the input timecode
  I_timecode* tc_in = NULL;
  // the output timecode
  I_timecode* tc_out = NULL;

  try
  {
    wcout << L"Retrieve MxfTk Error Handler" << endl;
    if (!GetMxfErrorHandlerInterface(&err_h))
      throw 1;

    wcout << L"MxfTk Initialization" << endl;
    InitMXF_TK();
    throw_on_error();

    wcout << L"Retrieve the frame rate" << endl;
    uint16_t frame_rate(0);
    stringstream fr_str;
    fr_str << argv[4];
    fr_str >> frame_rate;
    if (frame_rate==0 || frame_rate>100)
    {
      wcout << L"Invalid frame rate!" << endl;
    }

    wcout << L"Retrieve the drop frame property" << endl;
    bool drop_frame(false);
    if (argv[5] == string("yes"))
    {
      drop_frame = true;
    }
  
    wcout << L"Build TC IN" << endl;
    NewTimecodeInterface(&tc_in, argv[2], frame_rate, drop_frame);
    throw_on_error();

    wcout << L"Build TC OUT" << endl;
    NewTimecodeInterface(&tc_out, argv[3], frame_rate, drop_frame);
    throw_on_error();
 
    vector<string> lst_filenames;
    vector<I_mxf_file*> lst_mxf_files;
    partial_restore(argv[1], tc_in, tc_out, &lst_filenames, &lst_mxf_files);
    throw_on_error();

    wcout << L"Free partial restore files" << endl;
    for (vector<I_mxf_file*>::iterator it_mxf_files = lst_mxf_files.begin();
      it_mxf_files != lst_mxf_files.end();
      it_mxf_files++)
    {
      I_mxf_file* pi_mxf_file = *it_mxf_files;
      FreeMxfFileInterface(&pi_mxf_file);
    }
  }
  catch (int)
  {
    wcout << L"An error occured!" << std::endl;
    // Retrieving details of the error(s) and display it(them).
    display_errors();
  }

  wcout << L"Close Process" << endl;
  if (tc_in)
    FreeTimecodeInterface(&tc_in);
  if (tc_out)
    FreeTimecodeInterface(&tc_out);

  return 0;
}

I_timecode* get_tc_in(string file)
{
  I_mxf_file* pi_mxf_file(NULL);
  I_timecode* tc(NULL);

  try
  {
    NewMxfFileInterface(&pi_mxf_file, file.c_str(), CLOSED_METADATA_ONLY);
    throw_on_error();

    generic_list* gl = pi_mxf_file->output_material();
    if (gl)
    {
      gl->begin();
      I_generic_material* gmat = *gl->next();
      if (gmat)
      {
        I_timecode_material* tmat = gmat->get_timecode();
        if (tmat)
        {
          ordered_timecode_list* otl = tmat->get_timecode_list();
          if (otl)
          {
            otl->begin();
            I_timecode_component* tcomp = *otl->next();
            I_timecode* tc_tcomp = tcomp->get_timecode();
            CopyTimecodeInterface(&tc, tc_tcomp);
            throw_on_error();
            tmat->free_timecode_list(otl);
          }
        }
      }
      pi_mxf_file->free_generic_list(gl);
    }
  }
  catch(...)
  {
    
  }

  if (pi_mxf_file)
    FreeMxfFileInterface(&pi_mxf_file);

  return tc;
}

bool get_external_references(I_mxf_file* pi_mxf_file, vector<string>& lst)
{
  bool is_ext_mxf(true);
  lst.clear();

  concrete_list* cl = pi_mxf_file->embedded_material();
  if (cl)
  {
    I_concrete_material** pcmat = NULL;
    cl->begin();
    while ((pcmat = cl->next()) != NULL)
    {
      I_concrete_material* cmat = *pcmat;
      if (cmat)
      {
        track_list* tl = cmat->source_tracks();
        if (tl)
        {
          tl->begin();
          I_track** ptr = tl->next();
          I_concrete_track* ctr = (I_concrete_track*)*ptr;
          if (ctr)
          {
            I_essence_type* ess_t = ctr->get_essence_type();
            if (ess_t)
            {
              bool is_network_loc(false);
              locators* locs = ess_t->external_references(is_network_loc);
              if (!is_network_loc)
              {
                wcout << L"Don't know how to partial restore the ext ref." << endl;
                throw 1;
              }
              if (locs && locs->size()>0)
              {
                locs->begin();
                const char** path = locs->next();
                if (path && *path)
                {
                  string spath = *path;
                  size_t l = spath.length();
                  if (!(spath[l-4] == '.' && spath[l-3] == 'm' && spath[l-2] == 'x' && spath[l-1] == 'f')
                    && !(spath[l-4] == '.' && spath[l-3] == 'M' && spath[l-2] == 'X' && spath[l-1] == 'F')
                    )
                  {
                    // this is not an MXF file
                    is_ext_mxf = false;
                  }
                  lst.push_back(spath);
                }
              }
            }
          }
          cmat->free_tracks(tl);
        }
      }
    }
    pi_mxf_file->free_concrete_list(cl);
  }

  return is_ext_mxf;
}

bool partial_restore(const char* file, I_timecode* tc_in, I_timecode* tc_out, vector<string>* lst_files, vector<I_mxf_file*>* lst_mxf_files)
{
  bool ok(true);

  // the mxf file we are going to partial restore.
  I_mxf_file* pi_mxf_file = NULL;

  try
  {
    wcout << L"Create the mxf file partial restore interface for " << file << endl;
    NewPartialRestoreMxfFileInterface(&pi_mxf_file, file, tc_in, tc_out);
    throw_on_error();

    // build name of output file
    string out_filename = file;
    size_t off = out_filename.find_last_of(".");
    if (off==string::npos)
    {
      off = out_filename.find_last_of(".");
    }

    if (off!=string::npos)
    {
      out_filename.erase(off);
    }
    out_filename += "_";
    out_filename += tc_in->get_time();
    out_filename += "_to_";
    out_filename += tc_out->get_time();
    out_filename += ".mxf";
    for (size_t i = 2; i<out_filename.length(); i++)
    {
      if (out_filename[i] == ':')
      {
        out_filename[i] = '_';
      }
    }

    // Check if the file contains external references.
    bool partial_restore_done(false);
    if (pi_mxf_file->has_external_essence()) // case: ext ref to mxf file
    {
      wcout << L"File contains external references" << endl;
      // check operational pattern from source file!
      I_mxf_file* pi_mxf_file2(NULL);
      NewMxfFileInterface(&pi_mxf_file2, file, CLOSED_METADATA_ONLY);
      throw_on_error();

      // this program only support partial restore of Op1b MXF files.
      std::string op = pi_mxf_file2->operational_pattern();
      FreeMxfFileInterface(&pi_mxf_file2);
      if (op!="Op1b")
      {
        wcout << L"This program does not support partial restore of MXF files with ext. refs with operational pattern!=Op1b." << endl;
        throw 1;
      }

      vector<string> lst;
      bool ext_ref_to_mxf = get_external_references(pi_mxf_file, lst);
      if (ext_ref_to_mxf)
      {
        wcout << L"File contains external references to MXF files" << endl;
        FreeMxfFileInterface(&pi_mxf_file);
        throw_on_error();

        // retrieve tc in from original file.
        I_timecode* ori_tc_in = get_tc_in(file);
        if (!ori_tc_in)
        {
          wcout << L"Failed to retrieve start timecode" << endl;
          throw 1;
        }
        uint64_t fr_in = tc_in->get_frame_count() - ori_tc_in->get_frame_count();
        uint64_t fr_out = tc_out->get_frame_count() - ori_tc_in->get_frame_count();
        FreeTimecodeInterface(&ori_tc_in);

        // list of new ext ref files
        vector<string> lst_new_files;
        vector<I_mxf_file*> lst_new_mxf_files;

        // loop through each of the media files.
        for (vector<string>::iterator it = lst.begin(); it != lst.end(); it++)
        {
          string ext_ref_file = file;
          
          // build absolute path to ext ref file.
          size_t off = ext_ref_file.find_last_of("/");
          if (off==string::npos)
          {
            off = ext_ref_file.find_last_of("\\");
          }

          if (off==string::npos)
          {
            wcout << L"Failed to build path to external reference." << endl;
            throw 1;
          }
          ext_ref_file.erase(off);
          ext_ref_file += "/";
          ext_ref_file += *it; 

          // retrieve tc in from referenced file
          I_timecode* ext_tc_in = get_tc_in(ext_ref_file);
          if (!ext_tc_in)
          {
            wcout << L"Failed to retrieve start timecode" << endl;
            throw 1;
          }

          wcout << L"Build TC IN and TC OUT for external ref file" << endl;
          uint16_t fr = tc_in->get_frame_rate();
          uint16_t ext_fr = ext_tc_in->get_frame_rate();
          
          // Check the frame rates are related!
          if (fr != ext_fr)
          {
            wcout << L"Cannot proceed: The timecode track from referenced file does not have the same frame rate as the referencing file." << endl; 
          }

          uint64_t nb_fr_in = (fr_in*ext_fr)/fr;
          ext_tc_in->add(nb_fr_in);

          I_timecode* ext_tc_out = get_tc_in(ext_ref_file);
          uint64_t nb_fr_out = (fr_out*ext_fr)/fr;
          ext_tc_out->add(nb_fr_out);

          // partial restore ext ref file!
          partial_restore(ext_ref_file.c_str(), ext_tc_in, ext_tc_out, &lst_new_files, &lst_new_mxf_files);
          FreeTimecodeInterface(&ext_tc_in);
          FreeTimecodeInterface(&ext_tc_out);
          throw_on_error();
        }

        wcout << L"Build new Op1b file!" << endl;
        I_op1b_file* pi_op1b_file(NULL);
        NewOp1bFileInterface(&pi_op1b_file, file);
        throw_on_error();

        vector<string>::iterator it_filenames = lst_new_files.begin();
        for (vector<I_mxf_file*>::iterator it_mxf_files = lst_new_mxf_files.begin();
            it_mxf_files != lst_new_mxf_files.end();
            it_mxf_files++, it_filenames++)
        {
          string it_filename = *it_filenames;
          //build path relatively to new op1b file
          string outfiledir = out_filename;
          size_t off = outfiledir.find_last_of("/");
          if (off==string::npos)
          {
            off = outfiledir.find_last_of("\\");
          }

          if (off==string::npos)
          {
            wcout << L"Failed to build path to external reference." << endl;
            throw 1;
          }

          it_filename.erase(0, off+1);
          wchar_t* wit_filename = new wchar_t[it_filename.length()+1];
          wit_filename[it_filename.length()] = 0;
          iso7_to_utf16(wit_filename, it_filename.c_str(), (int)it_filename.length());

          // retrieve concrete material from new external file
          I_mxf_file* pi_ext_mxf_file = *it_mxf_files;
          concrete_list* cl = pi_ext_mxf_file->embedded_material();
          if (cl)
          {
            cl->begin();
            I_concrete_material* cmat = *cl->next();
            wcout << L"Append concrete material to Op1b file" << endl;
            pi_op1b_file->add_external_material(cmat, wit_filename);
            pi_ext_mxf_file->free_concrete_list(cl);
          }
          delete [] wit_filename;
          throw_on_error();
        }

        // set start timecode
        redefine_timecode(pi_op1b_file, tc_in);
        throw_on_error();

        wcout << L"Write mxf file on disk" << endl;
        pi_op1b_file->flush_file(out_filename.c_str());
        throw_on_error();

        wcout << L"Wait termination" << endl;
        pi_op1b_file->wait_end_flush_thread();
        throw_on_error();

        wcout << L"Free external files" << endl;
        for (vector<I_mxf_file*>::iterator it_mxf_files = lst_new_mxf_files.begin();
            it_mxf_files != lst_new_mxf_files.end();
            it_mxf_files++)
        {
          I_mxf_file* pi_ext_mxf_file = *it_mxf_files;
          FreeMxfFileInterface(&pi_ext_mxf_file);
        }
        partial_restore_done = true;
      }
      else
      {
        wcout << L"File contains external references to raw media files" << endl;
      }
    }

    if (!partial_restore_done) // case: no ext ref or ext ref to raw media
    {
      wcout << L"Write mxf file on disk" << endl;
      pi_mxf_file->flush_file(out_filename.c_str());
      throw_on_error();

      wcout << L"Wait termination" << endl;
      pi_mxf_file->wait_end_flush_thread();
      throw_on_error();
  
      if (lst_files)
      {
        lst_files->push_back(out_filename);
      }
      if (lst_mxf_files)
      {
        lst_mxf_files->push_back(pi_mxf_file);
      }
    }
  }
  catch (int)
  {
    wcout << L"An error occured!" << std::endl;
    // Retrieving details of the error(s) and display it(them).
    display_errors();
    ok = false;
  }

  return ok;
}

void redefine_timecode(I_op1b_file* pi_op1b_file, I_timecode* tc_in)
{
  // Output material
  I_generic_material* pi_gen_mat = NULL;

  // New timecode component (to add to timecode material) 
  I_timecode_component* pi_time_comp = NULL;

  // New timecode material
  I_timecode_material* pi_new_time_mat = NULL;
  bool delete_time_mat = false;

  try
  {
    wcout << L"\tGet output material (new timecode will be added to this material)" << endl;
    generic_list* gen_list = pi_op1b_file->output_material();
    gen_list->begin();
    pi_gen_mat = *gen_list->next();
    pi_op1b_file->free_generic_list(gen_list);
    throw_on_error();

    wcout << L"\tRetrieve output timecode that was automatically generated by MxfTk" << endl;
    I_timecode_material* pi_time_mat = pi_gen_mat->get_timecode();
    int64_t origin = pi_time_mat->get_origin();
    int64_t duration = pi_time_mat->get_total_duration();
    rational *edit_rate = pi_time_mat->get_edit_rate();
    throw_on_error();

    wcout << L"\tTimecode Redefinition" << endl;
    wcout << L"\tBuild new timecode material" << endl;
    NewTimecodeMaterialInterface(&pi_new_time_mat, origin, edit_rate);
    FreeRationalInterface(&edit_rate);
    throw_on_error();
    delete_time_mat = true;

    wcout << L"\tCreate new timecode component and append it to timecode material" << endl; 
    I_timecode* new_tc_in(NULL);
    CopyTimecodeInterface(&new_tc_in, tc_in);
    NewTimecodeComponentInterface(&pi_time_comp, new_tc_in, duration);
    pi_new_time_mat->append_timecode_component(pi_time_comp);
    throw_on_error();

    wcout << L"\tAttach New Timecode material to output material" << endl;
    pi_gen_mat->set_timecode(pi_new_time_mat); 
    throw_on_error();
    delete_time_mat = false; // pi_time_mat will be deleted upon destruction of pi_gen_mat
  }
  catch(int)
  {
    // do nothing will be rethrown in the following throw_on_error() 
    // and catched in main()
  }

  if (delete_time_mat)
    FreeTimecodeMaterialInterface(&pi_new_time_mat);

  throw_on_error();
}

void read_timecode(I_generic_material* pi_gen_mat)
{
  // Output timecode material
  I_timecode_material* pi_time_mat = NULL;

  // List of timecode components fro mthe timecode material
  ordered_timecode_list* time_list = NULL;

  try
  {
    pi_time_mat = pi_gen_mat->get_timecode();

    wstringstream str;
    rational *edit_rate = pi_time_mat->get_edit_rate();
    str << L"\t\tTimecode track: origin:" << (int)pi_time_mat->get_origin() << L"  duration:"  << (int)pi_time_mat->get_total_duration() << L"  edit rate:" << (int)edit_rate->get_num() << L"/" << (int)edit_rate->get_den() << endl;
    wcout << str.str(); 
    FreeRationalInterface(&edit_rate);
    throw_on_error();

    wcout << L"\t\tTimecode: " << endl;
    time_list = pi_time_mat->get_timecode_list();
    time_list->begin();
    I_timecode_component** tmc_p = NULL;
    while ((tmc_p = time_list->next())!=NULL)
    {
      I_timecode_component* tmc = *tmc_p;
      I_timecode* pi_tm = NULL;
      CopyTimecodeInterface(&pi_tm, tmc->get_timecode());	// copy it because we will change its value
      int64_t l_dur = tmc->get_duration();
      wcout << L"\t\t\t" << pi_tm->get_time() << L" until ";
      if (l_dur>=0)
      {
        pi_tm->add(l_dur-1);
        wcout << pi_tm->get_time() << endl;
      }
      else
      {
        wcout << L"??" << endl;
      }
      FreeTimecodeInterface(&pi_tm);
      throw_on_error();
    }  
  }
  catch(int)
  {
    // do nothing will be rethrown in the following throw_on_error() 
    // and catched in redefine_timecode()
  }

  if (time_list)
    pi_time_mat->free_timecode_list(time_list);

  throw_on_error();
}



void throw_on_error()
{
  if (err_h->has_an_error())
    throw 1;
}

void display_errors()
{
  // Retrieve error handler
  if (!err_h)
    return;

  // Retrieve list of errors
  error_list* err_l = err_h->get_errors();
  if (err_l)
  {
    err_l->begin();
    I_mxf_error** ppi_mxf_error = NULL;
    while ((ppi_mxf_error = err_l->next()) != NULL) //go through the list
    {
      I_mxf_error* pi_error = *ppi_mxf_error;
      // and display each error 
      wcout << L"Error " << (int)pi_error->get_id() << L": " << pi_error->get_error_msg() << endl;
    }
    // free error list interface
    err_h->free_errors(err_l);
  }
}
