//------------------------------------------------------------------------------
//  Author:    	MXF Team
//  Comment:	MXF_TK: This class defines the concrete_track interface.
//
//  The following convention is used for defining parameters and retrun values:
//      in : input parameter (will be used for computation in the function)
//      out : output parameter (will be modified by the function call)
//      in out : input parameter that may be modified by the function call.
//      free : MXFTk user is reponsible for the deletion of this parameter
//             unless it is passed as a parameter of another function without 
//             the "free" tag.
//
// Copyright (c) 2008 OPENCUBE. All rights reserved.
// -----------------------------------------------------------------------------
#ifndef OC_I_CONCRETE_TRACK_HPP
#define OC_I_CONCRETE_TRACK_HPP

#include <mxf_tk/mxf_tk.hpp>
#include <mxf_tk/I_track.hpp>
#include <mxf_tk/I_timecode.hpp>
#include <mxf_tk/I_metadata.hpp>
#include <mxf_tk/I_essence_type.hpp>

namespace opencube
{
  namespace MXFTK_NAMESPACE
  {
    /* I_concrete_track:
     * A class describing a track from a material.
     * This can be an essence or a descriptive metadata track.
     * Essence data can be read from this track
     */
    class I_concrete_track : public I_track
    {
    private:
      virtual bool CALLING_CONVENTION add_metadata(I_metadata_material*, track_list*)=0;
    public:
      /* 
       * Returns the descriptors of the essence stream of this track.
       * Returns NULL if not found.
       * Returned I_metadata interface shouldn't be destroyed by MXF_TK user.
       * Returns NULL if an error occured.
       */
      virtual metadata_list* /* free */ CALLING_CONVENTION get_descriptors() const=0;

      /* 
       * Free the descriptors list returned by get_descriptors()
       * Returns true if successful.
       */
      virtual bool CALLING_CONVENTION free_descriptors(metadata_list* /* in */)=0;

      /*
       * Returns a structure helping to identify the concrete track's content.
       * Returned pointer must not be deleted.
       */
      virtual I_essence_type* CALLING_CONVENTION get_essence_type()=0;

      /* 
       * Returns the locator of the essence stream of this track.
       * Returned string is useful while creating a mxf file. It enables you
       * to link each mxf track with the essence file it embeds.
       * Returned const char* shouldn't be destroyed by the MXF_TK user. It
       * will be destroyed upon deletion of this concrete_track.
       * Returns NULL if an error occured.
       */
      virtual const char* CALLING_CONVENTION get_locator() const=0;

      /*
       * Writes into a buffer the essence stream of this track.
       * Should be called only while reading an mxf file.
       * uint8_t* buffer to be filled by the essence (allocated by MXF_TK user)
       * size_t number of bytes to be written (buffer size)
       * Returns the number of bytes effectively written.
       */
      virtual size_t CALLING_CONVENTION read_eu(uint8_t* /* in out free */, 
                                                size_t s /* in */)=0;

      /* 
       * Seeks the track content in an absolute way, jumping to a given edit unit.
       * The int64_t parameter is the position along the track. Its value will
       * be added to the origin of this track. 
       * Therefore a track with an origin = 5 and a duration = 25 can be accessed
       * with a int64_t position comprised between -5 and 20.
       * If the seek operation cannot be performed (seeking beyond track
       * scope) returned value is 0xffffffffffffffff otherwise it returns the
       * position of the internal pointer.
       * Returns true if successful.
       */
      virtual bool CALLING_CONVENTION seek_eu(int64_t /* in */)=0;

      /*
       * Jumps to the next edit unit (if possible).
       * Returns true if successful.
       */
      virtual bool CALLING_CONVENTION seek_next_eu()=0;

      /*
       * Jumps to the previous edit unit (if possible). 
       * Returns true if successful.
       */
      virtual bool CALLING_CONVENTION seek_previous_eu()=0;

      /* 
       * Seeks the track content in an absolute way, jumping to a given timecode.
       * The current file must contain a timecode track (I_timecode_material).
       * We recommend usage of timecode for playout synchronization of different tracks.
       * If the seek operation cannot be performed (seeking beyond track
       * scope) returned value is false, true otherwise.
       */
      virtual bool CALLING_CONVENTION seek_timecode(const I_timecode* /* in free */)=0;

      /*
       * Returns how much data still requires to be read before reaching next edit unit.
       * You can use this function to read on edit unit by edit unit.
       * Returns 0 if all the data of this edit unit was already read or if an
       * error occured.
       */
      virtual uint64_t CALLING_CONVENTION current_eu_size()=0;

      /*
       * Returns the edit unit the internal pointer points to.
       * Returns 0x7fffffffffffffff if an error occured.
       */
      virtual int64_t CALLING_CONVENTION get_current_eu()=0;

      /*
       * Returns the current offset in bytes from the beginning of the file 
       * Returns 0xffffffffffffffff if an error occurred
       */
      virtual uint64_t CALLING_CONVENTION get_current_offset()=0;

      /**
       * Returns the cryptographic key id for this track, 
       * Return NULL if track is not crypted.
       * Returned pointer should not be destroyed.
       * If not NULL its size is 16
       */
      virtual const uint8_t * CALLING_CONVENTION get_cryptographic_key_id()=0;

      /**
       * Set the key used for descryption. Size of parameter must be 16
       * this key is retrieve from the cryptographic key id by the key manager.
       * This function has no effect on not crypted track
       */
      virtual void CALLING_CONVENTION set_cipher_key(const uint8_t * /* in free */)=0;

      /**
       * While reading an MXF file call this function to set the 
       * external ref file for this track.
       */
      virtual void CALLING_CONVENTION set_external_ref_file(const char * /* in free */)=0;
    };
  }
}

#endif
